from typing import List

from dt_shell.commands import DTCommandConfigurationAbs


class DTCommandConfiguration(DTCommandConfigurationAbs):

    @classmethod
    def aliases(cls) -> List[str]:
        return ["token"]



import json
import sys

from .duckietown_tokens import DuckietownToken


def verify_a_token_main(args=None):
    try:
        if args is None:
            args = sys.argv[1:]

        if args:
            token_s = args[0]
        else:
            msg = "Please enter token:\n> "
            token_s = input(msg)

        sys.stderr.write("Verifying token %r\n" % token_s)

        try:
            token = DuckietownToken.from_string(token_s)
        except ValueError:
            msg = "Invalid token format."
            sys.stderr.write(msg + "\n")
            sys.exit(3)

        o = dict()
        o["uid"] = token.uid
        o["payload"] = token.payload
        msg = json.dumps(o)
        print(msg)
        sys.exit(0)

    except Exception as e:
        sys.stderr.write(str(e) + "\n")
        sys.exit(3)



import dataclasses
import os
import traceback
from typing import Optional, List

import requests

from dt_shell.utils import pretty_json

import dt_shell

DTHUB_SCHEMA: str = "https"
DTHUB_HOST: str = os.environ.get("DTHUB_HOST", "hub.duckietown.com")
DTHUB_API_VERSION: str = "v1"
DTHUB_API_URL: str = f"{DTHUB_SCHEMA}://{DTHUB_HOST}/api/{DTHUB_API_VERSION}"


@dataclasses.dataclass
class HUBApiResponse:
    messages: List[str]
    result: dict
    code: int


class HUBApiError(Exception):

    def __init__(self, uri: str, response: dict):
        self.uri = uri
        assert not response["success"]
        self.messages: List[str] = response["messages"]
        self.result: dict = response["result"]
        self.code: int = response["code"]

    @property
    def human(self) -> str:
        parts: List[str] = [
            f"An error occurred while communicating with the API endpoint '{self.uri}' on the HUB."
        ]
        if self.code:
            parts.append(f"The error code returned is {self.code}.")
        if self.messages:
            messages: str = "The response contained the following messages:\n\t- " + \
                            "\n\t -".join(self.messages)
            parts.append(messages)
        if self.result:
            result: str = "The response contained the following result:\n" + \
                            pretty_json(self.result, indent_len=4)
            parts.append(result)
        # ---
        return "\n".join(parts)


def hub_api_post(endpoint: str, data: dict, token: Optional[str] = None) -> HUBApiResponse:
    # get token from the profile if not given explicitly
    if token is None:
        token = dt_shell.shell.profile.secrets.dt_token
    # compile url
    url: str = f"{DTHUB_API_URL}/{endpoint.lstrip('/')}"
    response: Optional[dict] = None
    try:
        response = requests.post(
            url,
            json=data,
            headers={"Authorization": f"Token {token}"}
        ).json()
        if not response["success"]:
            raise HUBApiError(endpoint, response)
        return HUBApiResponse(
            messages=response["messages"],
            result=response["result"],
            code=response["code"],
        )
    except HUBApiError as e:
        raise e
    except:
        raise HUBApiError(endpoint, {
            "success": False,
            "messages": [f"Python exception:\n{traceback.format_exc()}"],
            "result": response,
            "code": None,
        })



import os
import sys
from typing import Optional

# add the content of the environment variable EXTRA_PYTHONPATH to the current path
sys.path.extend(os.environ.get("EXTRA_PYTHONPATH", "").split(":"))

import logging

from dt_shell_cli import logger

from dt_shell import DTShell, dtslogger, CommandsLoadingException
from dt_shell.shell import get_cli_options
from dt_shell.logging import setup_logging_color, dts_print
from dt_shell.constants import DTShellConstants
from dt_shell.environments import Python3Environment
from dt_shell.checks.environment import abort_if_running_with_sudo


# NOTE: this file runs the shell in this interpreter and in quiet mode, the entrypoint should always be
#       the file dts.py.


def main() -> None:
    # make sure we are not running as sudo
    abort_if_running_with_sudo()

    # configure logger
    setup_logging_color()

    # parse shell options (anything between `dts` and the first word that does not start with --)
    cli_arguments = sys.argv[1:]
    cli_options, arguments = get_cli_options(cli_arguments)

    # propagate options to the constants
    DTShellConstants.DEBUG = cli_options.debug
    DTShellConstants.VERBOSE = cli_options.verbose
    DTShellConstants.QUIET = cli_options.quiet

    # process options here
    if cli_options.debug:
        dtslogger.setLevel(logging.DEBUG)

    # we run in quiet mode
    logger.setLevel(logging.WARNING)

    # instantiate shell
    shell: Optional[DTShell] = None
    try:
        shell = DTShell(
            skeleton=False,
            readonly=False,
            banner=False,
            billboard=False,
            profile=cli_options.profile
        )
    except CommandsLoadingException as e:
        dts_print("FATAL: " + str(e))
        exit(90)
    except Exception as e:
        dts_print("FATAL: " + str(e))
        exit(91)

    # run command in this interpreter
    Python3Environment().execute(shell, arguments)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import dataclasses
import datetime
import os
from typing import Optional, Dict, List

DEBUG = False
DNAME = "Duckietown Shell"
DEFAULT_ROOT = os.path.expanduser("~/.duckietown/shell/")
BASH_COMPLETION_DIR = os.path.expanduser("~/.local/share/bash-completion/completions")
DTHUB_URL = os.environ.get("DTHUB_URL", "https://hub.duckietown.com")


@dataclasses.dataclass
class Distro:
    name: str
    branch: str = None
    end_of_life: Optional[datetime.date] = None
    staging: bool = False
    stable: bool = False
    tokens_supported: List[str] = dataclasses.field(default_factory=list)
    token_preferred: Optional[str] = None

    def __post_init__(self):
        # if the branch is not given then it takes the distro name
        if self.branch is None:
            self.branch = self.name

    @property
    def end_of_life_fmt(self, fmt: str = "%d %B %Y") -> str:
        if self.end_of_life is None:
            return ""
        return self.end_of_life.strftime(fmt)

    def as_dict(self) -> dict:
        return dataclasses.asdict(self)


class DTShellConstants:
    PROFILE: Optional[str] = None
    ROOT: str = os.path.expanduser(os.environ.get("DTSHELL_ROOT", DEFAULT_ROOT))
    DEBUG: bool = False
    VERBOSE: bool = False
    QUIET: bool = False


# commands update
CHECK_CMDS_UPDATE_MINS = 5
CHECK_BILLBOARD_UPDATE_SECS = 60 * 60 * 24   # every 24 hours
PUSH_USER_EVENTS_TO_HUB_SECS = 60 * 60 * 1   # every 1 hour

SHELL_LIB_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULT_PROFILES_DIR = os.path.join(DEFAULT_ROOT, "profiles")

IGNORE_ENVIRONMENTS: bool = os.environ.get("IGNORE_ENVIRONMENTS", "0").lower() in ["1", "y", "yes"]

# distributions
KNOWN_DISTRIBUTIONS: Dict[str, Distro] = {
    # daffy
    "daffy": Distro(
        "daffy",
        "daffy",
        stable=True,
        # end_of_life=datetime.date(2024, 12, 31),
        tokens_supported=["dt1", "dt2"],
        token_preferred="dt1"
    ),
    "daffy-staging": Distro(
        "daffy",
        "daffy-staging",
        stable=True,
        # end_of_life=datetime.date(2024, 12, 31),
        staging=True,
        tokens_supported=["dt1", "dt2"],
        token_preferred="dt1"
    ),
    # ente
    "ente": Distro(
        "ente",
        "ente",
        stable=False,
        tokens_supported=["dt2"],
        token_preferred="dt2"
    ),
    "ente-staging": Distro(
        "ente",
        "ente-staging",
        stable=False,
        staging=True,
        tokens_supported=["dt2"],
        token_preferred="dt2"
    ),
}
SUGGESTED_DISTRIBUTION: str = "daffy"

# command set
EMBEDDED_COMMAND_SET_NAME: str = "embedded"
DEFAULT_COMMAND_SET_REPOSITORY = {
    "username": "duckietown",
    "project": "duckietown-shell-commands",
    "branch": SUGGESTED_DISTRIBUTION
}

# URLs
DUCKIETOWN_TOKEN_URL = "https://hub.duckietown.com/token"

# read requirements list embedded as asset into the release
SHELL_REQUIREMENTS_LIST: str = os.path.join(os.path.dirname(__file__), "assets", "requirements.txt")
assert os.path.exists(SHELL_REQUIREMENTS_LIST)

# database names
DB_PROFILES: str = "profiles"
DB_SETTINGS: str = "settings"
DB_SECRETS: str = "secrets"
DB_SECRETS_DOCKER: str = "secrets_docker"
DB_COMMAND_SET_UPDATES_CHECK: str = "command_sets_updates_check"
DB_INSTALLED_DEPENDENCIES: str = "installed_dependencies"
DB_USER_COMMAND_SETS_REPOSITORIES: str = "user_command_sets_repositories"
DB_MIGRATIONS: str = "migrations"
DB_UPDATES_CHECK: str = "updates_check"
DB_BILLBOARDS: str = "billboards"
DB_STATISTICS_EVENTS: str = "stats_events"



import sys
from typing import Optional, List, Any, Union

from . import __version__


class NotFound(Exception):
    pass


class InvalidEnvironment(Exception):
    pass


class InvalidRemote(Exception):
    pass


class UserError(Exception):
    """an error that will be briefly printed"""
    pass


class UserAborted(KeyboardInterrupt):
    """the user interrupted the execution"""
    pass


class InvalidConfig(Exception):
    pass


class CommandsLoadingException(Exception):
    pass


class ConfigInvalid(Exception):
    pass


class ConfigNotPresent(Exception):
    pass


class CouldNotGetVersion(Exception):
    pass


class URLException(Exception):
    pass


class NoCacheAvailable(Exception):
    pass


class ShellNeedsUpdate(Exception):
    def __init__(self, needed: str):
        self._version_needed: str = needed
        self._current_version: str = __version__

    @property
    def current_version(self) -> str:
        return self._current_version

    @property
    def version_needed(self) -> str:
        return self._version_needed

    @staticmethod
    def assert_newer_or_equal_to(needed: str):
        from .utils import parse_version

        exc = ShellNeedsUpdate(needed)
        vnow = parse_version(exc.current_version)
        vneed = parse_version(needed)
        if vneed > vnow:
            raise exc


class ShellInitException(Exception):

    def __init__(self,
                 msg: str,
                 stdout: Optional[Union[str, bytes]] = None,
                 stderr: Optional[Union[str, bytes]] = None):
        from dt_shell.logging import dts_print
        # write stdout
        if stdout:
            if isinstance(stdout, bytes):
                stdout = stdout.decode("utf-8")
            dts_print(stdout, color="red")
            sys.stdout.flush()
        # write stderr
        if stderr:
            if isinstance(stderr, bytes):
                stderr = stderr.decode("utf-8")
            sys.stderr.write(stderr)
            sys.stderr.flush()
        # store message
        super(ShellInitException, self).__init__(msg)


class CommandNotFound(Exception):

    def __init__(self, last_matched: Optional[Any], remaining: List[str]):
        from dt_shell import DTCommandAbs
        self.last_matched: Optional[DTCommandAbs] = last_matched
        self.remaining: List[str] = remaining


class RunCommandException(RuntimeError):

    def __init__(self, msg: str, exit_code: int, stdout: str, stderr: str):
        self.msg: str = msg
        self.exit_code: int = exit_code
        self.stdout: str = stdout
        self.stderr: str = stderr
        super(RunCommandException, self).__init__(msg)



# This is a default __init__ file for the Duckietown Shell commands
#
# Maintainer: Andrea F. Daniele

# import current command
try:
    # noinspection PyUnresolvedReferences
    from .command import DTCommand

except ImportError:
    pass

import glob

# import subcommands
from os.path import basename, dirname, isdir

modules = glob.glob(dirname(__file__) + "/*")

# load submodules
for mod in [m for m in modules if isdir(m)]:
    try:
        exec("import %s" % basename(mod))
    except ImportError:
        pass



from typing import Optional, Set, List

from dt_shell import DTCommandAbs, DTShell
from dt_shell.constants import Distro, KNOWN_DISTRIBUTIONS
from dt_shell.profile import ShellProfile
from utils.table_utils import format_matrix


class DTCommand(DTCommandAbs):
    help = 'List all known profiles'

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # get list of existing profiles
        profiles: Set[str] = set(shell.profiles.keys())
        # make a table
        header = ["Profile", "Distribution", "Staging"]
        data = []
        for profile_name in sorted(profiles):
            profile: ShellProfile = ShellProfile(profile_name, readonly=True)
            # find distro
            distro: Optional[Distro] = None
            for d in KNOWN_DISTRIBUTIONS.values():
                if d.name == profile.distro.name:
                    distro = d
                    break
            # current?
            current: str = ">>" if shell.profile.name == profile_name else ""
            # add to table
            data.append([
                current,
                profile_name,
                profile.distro.name,
                ("Yes" if distro.staging else "No") if distro else "NA"
            ])
        # render table
        print()
        print(format_matrix(header, data, "{:^{}}", "{:<{}}", "{:>{}}", "\n", " | "))
        print()

    @staticmethod
    def complete(shell: DTShell, word: str, line: str) -> List[str]:
        return []



import logging
import os
import sys
from typing import Optional, Dict, List

# NOTE: DO NOT IMPORT DT_SHELL HERE

from . import logger


# noinspection PyPep8Naming
def dts():
    # make sure we have not imported dt_shell yet
    modules = [m.__name__ for m in sys.modules.values() if m]
    if "dt_shell" in modules:
        logger.fatal("The module 'dt_shell' was found imported before we had a chance to update "
                     "the PYTHONPATH. This should not have happened. Please, contact technical support.")
        return

    # custom path to dt_shell library can be set using the DTSHELL_LIB environment variable
    DTSHELL_LIB = os.environ.get("DTSHELL_LIB", None)
    if DTSHELL_LIB:
        DTSHELL_LIB = os.path.abspath(DTSHELL_LIB)
        # make sure the duckietown_shell library exists in the given path
        dt_shell_dir = os.path.join(DTSHELL_LIB, "dt_shell")
        if not os.path.exists(dt_shell_dir) or not os.path.isdir(dt_shell_dir):
            logger.fatal("Duckietown Shell library not found in the given DTSHELL_LIB path. "
                         f"Directory '{dt_shell_dir}' does not exist.\n")
            sys.exit(1)
        # make sure the duckietown_shell library is a Python package
        dt_shell_init = os.path.join(DTSHELL_LIB, "dt_shell", "__init__.py")
        if not os.path.exists(dt_shell_init) or not os.path.isfile(dt_shell_init):
            logger.fatal(f"The given directory '{dt_shell_dir}' is not a Python package.\n")
            sys.exit(2)
        # give the given path the highest priority
        sys.path.insert(0, DTSHELL_LIB)

    # import dt_shell
    import dt_shell
    from dt_shell.constants import DTShellConstants
    from dt_shell.logging import setup_logging_color, dts_print
    from dt_shell.checks.environment import abort_if_running_with_sudo
    from dt_shell.shell import get_cli_options
    from dt_shell.commands import CommandDescriptor
    from dt_shell.environments import ShellCommandEnvironmentAbs
    from dt_shell.exceptions import CommandNotFound, ShellInitException, UserAborted, UserError
    from dt_shell.utils import replace_spaces, print_debug_info
    from dt_shell import DTShell, dtslogger

    # parse shell options (anything between `dts` and the first word that does not start with --)
    cli_arguments = sys.argv[1:]
    cli_options, arguments = get_cli_options(cli_arguments)

    # perform complete
    if cli_options.complete:
        # disable loggers
        # NOTE: comment these two lines if you want to test what happens when --complete is run. You might
        #       discover that some things that should not happen in readonly mode do indeed happen.
        #       See DTSW-4277 for an example.
        dtslogger.setLevel(logging.CRITICAL + 1)
        logger.setLevel(logging.CRITICAL + 1)
        # ---
        complete()
        exit()

    # propagate options to the constants
    DTShellConstants.DEBUG = cli_options.debug
    DTShellConstants.VERBOSE = cli_options.verbose
    DTShellConstants.QUIET = cli_options.quiet

    # notify user of their choices
    if DTSHELL_LIB:
        logger.info(f"Using duckietown-shell library from '{DTSHELL_LIB}' as instructed by the environment "
                    f"variable DTSHELL_LIB.")
    if DTShellConstants.VERBOSE:
        logger.info(f"Using duckietown-shell library from '{dt_shell.__file__}'")

    # make sure we are not running as sudo
    abort_if_running_with_sudo()

    # configure logger
    setup_logging_color()

    # process options here
    if DTShellConstants.DEBUG:
        dtslogger.setLevel(logging.DEBUG)
    if DTShellConstants.VERBOSE:
        logger.setLevel(logging.DEBUG)

    # load shell skeleton
    try:
        shell = DTShell(
            skeleton=True,
            readonly=False,
            banner=True,
            billboard=True,
            profile=cli_options.profile
        )
    except (UserAborted, KeyboardInterrupt):
        dts_print("User aborted operation.")
        return
    except UserError as e:
        msg = str(e)
        dts_print(msg, "red")
        print_debug_info()
        sys.exit(1)

    # if we don't have a profile, we bail
    if shell.profile is None:
        raise RuntimeError("The shell could not load a profile. This should not have happened, please "
                           "contact technical support")
        # TODO: maybe suggest clearing the profile directory?

    # get command's environment and use it to execute the command
    arguments = list(map(replace_spaces, arguments))
    cmdline = " ".join(arguments)
    command: Optional[CommandDescriptor] = None
    try:
        command = shell.get_command(cmdline)
    except CommandNotFound as e:
        inpt: str = cmdline.strip()
        if e.last_matched is None:
            if len(inpt) <= 0:
                # no input, but we only complain if the experience was not interactive
                if shell.performed_migrations or shell.configured_shell or shell.configured_profile:
                    exit(0)
                # no input
                # TODO: maybe suggest possible commands?
                dts_print("Use the syntax\n\n"
                          "\t\tdts [options] command [subcommand1 [subcommand2] ...] [arguments]\n",
                          color="red")
                exit(1)
            else:
                # input was given but it was not recognized
                shell.default(cmdline)

        else:
            # we have a partial match of the arguments
            word: Optional[str] = e.remaining[0] if e.remaining else None
            subcommands: Dict[str] = e.last_matched.commands
            if len(subcommands) > 0:
                subcommands_list: str = "\n\t\t".join(subcommands.keys())
                # the partially matched command has subcommands
                if word:
                    dts_print(
                        f"Sub-command '{word}' not recognized.\n"
                        f"Available sub-commands are:\n\n\t\t{subcommands_list}"
                    )
                else:
                    dts_print(f"Available sub-commands are:\n\n\t\t{subcommands_list}")
            else:
                # the partially matched command is a leaf command (i.e., no subcommands)
                # TODO: make sure this does not happen
                raise NotImplementedError("NOT IMPLEMENTED")

    if command is not None:
        env: ShellCommandEnvironmentAbs = command.environment
        logger.debug(f"Running command '{command.selector}' in environment '{env.__class__.__name__}'")
        try:
            env.execute(shell, arguments)
        except ShellInitException:
            logger.error("An error occurred, the reason for the error should be printed above.")
            exit(99)


def complete():
    """
    NOTE: If you want to test the autocomplete output, use the following command:
        .
            dts --complete <idx> dts <word1> <word2> ...
        .
        where <idx> is the index corresponding to the word we want to complete.
        For example, pressing <Tab> on the partial command "dts devel bu" would generate the
        command
        .
            dts --complete 2 dts devel bu
        .
        which, if run, would print the string "bump build" for BASH to break at the space.
    """
    from dt_shell import DTShell

    try:
        shell = DTShell(
            skeleton=True,
            readonly=True,
            banner=False,
            billboard=False
        )
    except:
        exit()

    def do_complete(comp_cword: str, *comp_words: str):
        comp_cword: int = int(comp_cword)
        comp_words: List[str] = list(comp_words)
        # add empty word if the pointer is past the last word (we are list all possible next words)
        if comp_cword == len(comp_words):
            comp_words.append("")
        # ---
        comp_line: str = " ".join(comp_words[1:])
        comp_word: str = comp_words[comp_cword]
        root_cmd: str = comp_words[1]
        if root_cmd in shell.commands:
            complete_fcn = getattr(shell, f"complete_{root_cmd}")
            return complete_fcn(comp_word, comp_line, 0, 0)
        else:
            items = shell.commands.keys()
            return (item for item in items if item.startswith(comp_word))

    suggestions: List[str] = do_complete(*sys.argv[2:])
    if suggestions:
        sys.stdout.write(" ".join(suggestions))
        sys.stdout.flush()
    exit(0)


if __name__ == '__main__':
    dts()



from typing import List

from dt_shell.constants import EMBEDDED_COMMAND_SET_NAME

from dt_shell import DTCommandAbs, DTShell


class DTCommand(DTCommandAbs):
    help = "Shows the list of all the commands available in the shell."

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # show core commands
        print("\nCore commands:")
        for cmd in shell.command_set(EMBEDDED_COMMAND_SET_NAME).commands.keys():
            print("\t%s" % cmd)

        # show commands grouped by command sets
        for cs in shell.command_sets:
            if cs.name == EMBEDDED_COMMAND_SET_NAME:
                continue
            print(f"\nCommand set '{cs.name}':")
            for cmd in cs.commands.keys():
                print("\t%s" % cmd)

    @staticmethod
    def complete(shell: DTShell, word: str, line: str) -> List[str]:
        return []



# This replaces the old default __init__ file for the Duckietown Shell commands
#
# Maintainer: Andrea F. Daniele
import copy
import importlib
import os.path
import sys
from os.path import (
    exists as _exists,
    dirname as _dirname,
    relpath as _relpath,
    join as _join,
)
from typing import Type, List

from dt_shell.constants import DTShellConstants

from .. import logger
from ..exceptions import ShellNeedsUpdate, CommandsLoadingException
from .commands import default_command_configuration, failed_to_load_command, DTCommandConfigurationAbs, \
    DTCommandAbs, CommandSet, DTCommandSetConfigurationAbs, default_commandset_configuration


def import_commandset_configuration(command_set: CommandSet) -> Type[DTCommandSetConfigurationAbs]:
    # constants
    _configuration_file = _join(command_set.path, "__command_set__", "configuration.py")
    # import command set configuration
    if _exists(_configuration_file):
        if DTShellConstants.VERBOSE:
            logger.debug(f"Importing configuration for command set '{command_set.name}' from "
                         f"'{_configuration_file}'")
        _configuration_sel: str = "__command_set__.configuration"
        # we temporarily add the path to the command set to PYTHONPATH
        old: List[str] = copy.deepcopy(sys.path)
        sys.path.insert(0, os.path.abspath(command_set.path))
        try:
            # import/refresh the configuration module
            configuration = importlib.import_module(_configuration_sel)
            if configuration.__file__ != _configuration_file:
                importlib.reload(configuration)
            # make sure we got the right one
            assert configuration.__file__ == _configuration_file
        finally:
            # restore PYTHONPATH
            sys.path = old

        DTCommandSetConfiguration: Type[DTCommandSetConfigurationAbs] = \
            configuration.DTCommandSetConfiguration
        if not issubclass(DTCommandSetConfiguration.__class__, DTCommandSetConfigurationAbs.__class__):
            msg = f"Cannot load command set configuration class in {_configuration_file}, the class " \
                  f"'DTCommandSetConfiguration' must extend the class 'DTCommandSetConfigurationAbs'"
            raise CommandsLoadingException(msg)
        # populate path
        DTCommandSetConfiguration.path = command_set.path

        return DTCommandSetConfiguration
    else:
        return default_commandset_configuration.DTCommandSetConfiguration


def import_commandset_init(command_set: CommandSet):
    # constants
    _init_file = _join(command_set.path, "__command_set__", "init.py")
    # skip if there is no init file for this command set
    if not _exists(_init_file):
        return
    # import command set init
    if DTShellConstants.VERBOSE:
        logger.debug(f"Executing init script for command set '{command_set.name}' from '{_init_file}'")
    _init_sel: str = "__command_set__.init"
    # we temporarily add the path to the command set to PYTHONPATH
    old: List[str] = copy.deepcopy(sys.path)
    sys.path.insert(0, os.path.abspath(command_set.path))
    # import/refresh the init module
    try:
        init = importlib.import_module(_init_sel)
        if init.__file__ != _init_file:
            importlib.reload(init)
        # make sure we got the right one
        assert init.__file__ == _init_file
    finally:
        # restore PYTHONPATH
        sys.path = old


def import_configuration(command_set: CommandSet, selector: str) -> Type[DTCommandConfigurationAbs]:
    # constants
    _command_dir = command_set.command_path(selector)
    _configuration_file = _join(_command_dir, "configuration.py")
    # import command configuration
    if _exists(_configuration_file):
        _command_sel: str = _relpath(_command_dir, command_set.path).strip("/").replace("/", ".")
        _configuration_sel: str = f"{_command_sel}.configuration"
        # we temporarily add the path to the command set to PYTHONPATH
        old: List[str] = copy.deepcopy(sys.path)
        sys.path.insert(0, os.path.abspath(command_set.path))
        try:
            if DTShellConstants.VERBOSE:
                logger.debug(f"Importing configuration for command '{_command_sel}' from "
                             f"'{_configuration_file}'")
            configuration = importlib.import_module(_configuration_sel)
        except ShellNeedsUpdate as e:
            logger.warning(
                f"Command '{_command_sel}' was not loaded because the shell needs to be "
                f"updated. Current version is {e.current_version}, required version is "
                f"{e.version_needed}"
            )
            return default_command_configuration.DTCommandConfiguration
        finally:
            # restore PYTHONPATH
            sys.path = old

        DTCommandConfiguration = configuration.DTCommandConfiguration
        if not issubclass(DTCommandConfiguration.__class__, DTCommandConfigurationAbs.__class__):
            msg = f"Cannot load command configuration class in {_configuration_file}, the class " \
                  f"'DTCommandConfiguration' must extend the class 'DTCommandConfigurationAbs'"
            raise CommandsLoadingException(msg)

        return DTCommandConfiguration
    else:
        return default_command_configuration.DTCommandConfiguration


def import_command(command_set: CommandSet, fpath: str) -> Type[DTCommandAbs]:
    # we always start from an __init__.py file
    if os.path.isdir(fpath):
        fpath = os.path.join(fpath, "__init__.py")
    # constants
    _command_dir = _dirname(fpath) if os.path.isfile(fpath) else fpath
    _command_file = _join(_command_dir, "command.py")
    # import current command
    if _exists(_command_file):
        _command_sel: str = _relpath(_command_dir, command_set.path).strip("/").replace("/", ".")
        _command_sel: str = f"{_command_sel}.command"
        try:
            if DTShellConstants.VERBOSE:
                logger.debug(f"Importing command '{_command_sel}' from '{_dirname(fpath)}/'")
            command = importlib.import_module(_command_sel)
        except ShellNeedsUpdate as e:
            logger.warning(
                f"Command '{_command_sel}' was not loaded because the shell needs to be "
                f"updated. Current version is {e.current_version}, required version is "
                f"{e.version_needed}"
            )
            return failed_to_load_command.DTCommand

        DTCommand = command.DTCommand        # handle wrong class
        if not issubclass(DTCommand.__class__, DTCommandAbs.__class__):
            msg = f"Cannot load command class in {_command_file}, the class " \
                  f"'DTCommand' must extend the class 'DTCommandAbs'"
            raise CommandsLoadingException(msg)
        return DTCommand
    else:
        return failed_to_load_command.DTCommand



import dataclasses
import os
import time
from typing import Iterator, Optional

from .constants import DB_STATISTICS_EVENTS
from .database import DTShellDatabase


@dataclasses.dataclass
class StatsEvent:
    name: str
    time: float
    payload: dict
    __db__: 'ShellProfileEventsDatabase'
    __key__: str
    # optional
    format: int = 1
    labels: Optional[dict] = None

    @property
    def time_millis(self) -> int:
        return int(self.time * 1000)

    def delete(self):
        self.__db__.delete(self.__key__)


class ShellProfileEventsDatabase(DTShellDatabase[dict]):

    @classmethod
    def load(cls, location: str):
        return ShellProfileEventsDatabase.open(DB_STATISTICS_EVENTS, location=location)

    @property
    def in_memory(self) -> bool:
        return os.environ.get("DTSHELL_DISABLE_STATS", "0").lower() in ["1", "yes", "true"]

    def get(self, *_, **__):
        raise NotImplementedError("Use the method ShellProfileEventsDatabase.events() instead.")

    def set(self, *_, **__):
        raise NotImplementedError("Use the method ShellProfileEventsDatabase.new() instead.")

    def events(self) -> Iterator[StatsEvent]:
        for key in list(self.keys()):
            value: dict = super(ShellProfileEventsDatabase, self).get(key)
            try:
                yield StatsEvent(**value, __db__=self, __key__=key)
            except TypeError:
                self.delete(key)

    def new(self, name: str, payload: dict = None, when: float = None, format: int = 1,
            labels: dict = None) -> StatsEvent:
        now: float = time.time()
        key: str = str(now)
        value: dict = {
            "name": name,
            "time": when or now,
            "payload": payload or {},
            "format": format,
            "labels": labels
        }
        if not self.in_memory:
            super(ShellProfileEventsDatabase, self).set(key, value)
        return StatsEvent(**value, __db__=self, __key__=key)



from typing import List

from dt_authentication import DuckietownToken
from dt_shell import DTCommandAbs, DTShell


class DTCommand(DTCommandAbs):
    @staticmethod
    def command(shell: DTShell, args: List[str]):
        token: str = shell.profile.secrets.dt_token
        # show token info
        uid = DuckietownToken.from_string(token).uid
        shell.sprint("Correctly identified as uid = %s" % uid)



import atexit
import copy
import dataclasses
import json
import os.path
import tempfile
import time
from typing import Optional, List, Dict, Union, Iterator, Tuple, Type

import questionary
from questionary import Choice

from dt_authentication import DuckietownToken
from . import logger, __version__
from .commands import CommandSet, CommandDescriptor
from .commands.repository import CommandsRepository
from .constants import DUCKIETOWN_TOKEN_URL, SHELL_LIB_DIR, DEFAULT_COMMAND_SET_REPOSITORY, \
    DEFAULT_PROFILES_DIR, DB_SECRETS, DB_SECRETS_DOCKER, DB_SETTINGS, DB_USER_COMMAND_SETS_REPOSITORIES, \
    DB_PROFILES, KNOWN_DISTRIBUTIONS, SUGGESTED_DISTRIBUTION, DB_UPDATES_CHECK, EMBEDDED_COMMAND_SET_NAME, \
    Distro
from .database.database import DTShellDatabase, NOTSET, DTSerializable
from .statistics import ShellProfileEventsDatabase
from .utils import safe_pathname, validator_token, yellow_bold, cli_style, parse_version, render_version, \
    indent_block, DebugInfo
from .exceptions import ConfigNotPresent

TupleVersion = Tuple[int, int, int]


@dataclasses.dataclass
class GenericCredentials(DTSerializable):

    username: str
    password: str

    @property
    def secret(self) -> str:
        return self.password

    def dump(self) -> dict:
        return {"username": self.username, "password": self.password}

    @classmethod
    def load(cls, v: dict) -> 'GenericCredentials':
        return GenericCredentials(**v)


class DockerCredentials(DTShellDatabase[dict]):

    @classmethod
    def load(cls, location: str):
        return DockerCredentials.open(DB_SECRETS_DOCKER, location=location)

    def contains(self, registry: str) -> bool:
        return super(DockerCredentials, self).contains(registry)

    def get(self, registry: str, default: GenericCredentials = NOTSET) -> GenericCredentials:
        try:
            d: dict = super(DockerCredentials, self).get(key=registry)
            return GenericCredentials.load(d)
        except DTShellDatabase.NotFound:
            if default is NOTSET:
                raise DTShellDatabase.NotFound(f"No credentials found for registry '{registry}' in database.")
            else:
                return default

    def set(self, registry: str, value: GenericCredentials):
        super(DockerCredentials, self).set(registry, value.dump())


class ShellProfileSecrets(DTShellDatabase):

    def __new__(cls, *args, **kwargs):
        inst = super().__new__(cls)
        inst._profile = kwargs["profile"]
        return inst

    @classmethod
    def load(cls, profile: 'ShellProfile', location: str):
        return ShellProfileSecrets.open(DB_SECRETS, location=location, init_args={"profile": profile})

    @property
    def dt_token(self) -> Optional[str]:
        preferred: str = self._profile.distro.token_preferred
        if preferred == "dt1":
            return self.dt1_token
        elif preferred == "dt2":
            return self.dt2_token
        else:
            raise ValueError(f"Token version '{preferred}' not supported")

    @dt_token.setter
    def dt_token(self, value: str):
        preferred: str = self._profile.distro.token_preferred
        if preferred == "dt1":
            self.dt1_token = value
        elif preferred == "dt2":
            self.dt2_token = value
        else:
            raise ValueError(f"Token version '{preferred}' not supported")

    @property
    def dt1_token(self) -> Optional[str]:
        return self.get("token/dt1", None)

    @dt1_token.setter
    def dt1_token(self, value: str):
        assert isinstance(value, str)
        self.set("token/dt1", value)

    @property
    def dt2_token(self) -> Optional[str]:
        return self.get("token/dt2", None)

    @dt2_token.setter
    def dt2_token(self, value: str):
        assert isinstance(value, str)
        self.set("token/dt2", value)

    @property
    def docker_credentials(self) -> DockerCredentials:
        return DockerCredentials.load(self._location)

    @classmethod
    def _as_temporary_file(cls, content: str, autoremove: bool = True) -> str:
        # create temporary file and write content to it
        tmpf_fd, tmpf_fpath = tempfile.mkstemp()
        os.write(tmpf_fd, content.encode("utf-8"))
        os.close(tmpf_fd)
        # when Python exits, the file gets removed
        if autoremove:
            atexit.register(os.remove, tmpf_fpath)
        # ---
        return tmpf_fpath

    def as_temporary_json_file(self, key: str, autoremove: bool = True) -> str:
        secret: dict = self.get(key)
        return self._as_temporary_file(json.dumps(secret), autoremove=autoremove)


class ShellProfileSettings(DTShellDatabase):

    @classmethod
    def load(cls, location: str):
        return ShellProfileSettings.open(DB_SETTINGS, location=location)

    @property
    def distro(self) -> Optional[str]:
        return self.get("distro", None)

    @distro.setter
    def distro(self, value: str):
        assert isinstance(value, str)
        self.set("distro", value)

    @property
    def check_for_updates(self) -> bool:
        return self.get("check_for_updates", True)

    @check_for_updates.setter
    def check_for_updates(self, value: bool):
        assert isinstance(value, bool)
        self.set("check_for_updates", value)

    @property
    def auto_update(self) -> bool:
        return self.get("auto_update", True)

    @auto_update.setter
    def auto_update(self, value: bool):
        assert isinstance(value, bool)
        self.set("auto_update", value)


@dataclasses.dataclass
class ShellProfile:
    name: str
    path: Optional[str] = None
    command_sets: List[CommandSet] = dataclasses.field(default_factory=list)
    readonly: bool = False

    _distro: dataclasses.InitVar[str] = None

    def __post_init__(self, _distro: Optional[str] = None):
        # load from disk
        if self.path is None:
            profiles_dir: str = os.environ.get("DTSHELL_PROFILES", DEFAULT_PROFILES_DIR)
            if profiles_dir != DEFAULT_PROFILES_DIR:
                logger.info(f"Loading profiles from '{profiles_dir}' as prescribed by the environment "
                            f"variable DTSHELL_PROFILES.")
            safe_name: str = safe_pathname(self.name)
            self.path = os.path.join(profiles_dir, safe_name)
            # make sure the profile directory exists
            if not os.path.exists(self.path) and not self.readonly:
                os.makedirs(self.path)

        # add profile to database
        if not self.readonly:
            # record new profile
            db: DTShellDatabase = DTShellDatabase.open(DB_PROFILES)
            if not db.contains(self.name):
                db.set(self.name, self.path)

        # updates check database
        self.updates_check_db: DTShellDatabase[float] = self.database(DB_UPDATES_CHECK)

        # set distro if given
        if _distro is not None:
            self.distro = _distro

        # this is the order with which command sets are loaded
        self.command_sets: List[CommandSet] = []

        # load command sets
        if "DTSHELL_COMMANDS" in os.environ:
            commands_path = os.environ["DTSHELL_COMMANDS"]
            # make sure the given path exists
            if not os.path.exists(commands_path):
                msg = f"The path {commands_path} given with the environment variable DTSHELL_COMMANDS does " \
                      f"not exist."
                raise Exception(msg)
            # load commands from given path
            msg = f"Loading commands from '{commands_path}' as instructed by the environment variable " \
                  f"DTSHELL_COMMANDS."
            logger.info(msg)
            self.command_sets.append(
                CommandSet(
                    "development",
                    commands_path,
                    profile=self,
                    repository=CommandsRepository.from_file_system(commands_path),
                    leave_alone=True,
                )
            )
            # TODO: this is where we update the profile.distro by taking the branch from the repository but
            #  only for this session
        else:
            profile_command_sets_dir: str = os.path.join(self.path, "commands")
            # add the default 'duckietown' command set
            if self.distro is not None:
                repository: CommandsRepository = CommandsRepository(
                    **{**DEFAULT_COMMAND_SET_REPOSITORY, "branch": self.distro.branch}
                )
                self.command_sets.append(
                    CommandSet(
                        "duckietown",
                        os.path.join(profile_command_sets_dir, "duckietown"),
                        profile=self,
                        repository=repository,
                    )
                )

            # add user defined command sets
            for n, r in self.user_command_sets_repositories:
                self.command_sets.append(CommandSet(n, os.path.join(profile_command_sets_dir, n), self, r))

        # we always add the embedded command set last so that it can override everything the others do
        self.command_sets.append(
            CommandSet(
                EMBEDDED_COMMAND_SET_NAME,
                os.path.join(SHELL_LIB_DIR, "embedded"),
                profile=self,
                leave_alone=True,
            )
        )

        # add command set versions to debugging data
        for cs in self.command_sets:
            DebugInfo.name2versions[f"command_set/{cs.name}"] = cs.version

        # drop all the command sets that do not support this version of the shell
        for cs in copy.copy(self.command_sets):
            vnow: TupleVersion = parse_version(__version__)
            vmin: Optional[TupleVersion] = cs.configuration.minimum_shell_version()
            vmax: Optional[TupleVersion] = cs.configuration.maximum_shell_version()
            # check min
            if vmin is not None and vnow < vmin:
                logger.warning(f"\n -- WARNING\n\n" +
                                indent_block(
                                   f"Command set '{cs.name}' wants a Duckietown Shell v{render_version(vmin)}"
                                   f" or newer. We are running v{render_version(vnow)}.\n"
                                   f"You will need to upgrade your shell to be able to use this command "
                                   f"set or switch to an older version of this command set.\n"
                                   f"This command set will now be disabled."
                                ) +
                               f"\n\n -- WARNING\n")
                self.command_sets.remove(cs)
            # check max
            if vmax is not None and vnow > vmax:
                logger.warning(f"\n -- WARNING\n\n" +
                               indent_block(
                                   f"Command set '{cs.name}' only supports Duckietown Shell up to "
                                   f"v{render_version(vmax)}. We are running v{render_version(vnow)}.\n"
                                   f"You will need to downgrade your shell to be able to use this command "
                                   f"set or switch to a newer version of this command set.\n"
                                   f"This command set will now be disabled."
                               ) +
                               f"\n\n -- WARNING\n")
                self.command_sets.remove(cs)

    @property
    def commands(self) -> Dict[str, Union[dict, CommandDescriptor]]:
        # collapse all command sets into a single tree of commands
        cmds = {}
        for cmd_set in self.command_sets:
            cmds.update(cmd_set.commands)
        return cmds

    @property
    def _databases_location(self) -> str:
        return os.path.join(self.path, "databases")

    @property
    def user_command_sets_repositories(self) -> Iterator[Tuple[str, CommandsRepository]]:
        return self.database(DB_USER_COMMAND_SETS_REPOSITORIES).items()

    @property
    def settings(self) -> ShellProfileSettings:
        return ShellProfileSettings.load(location=self._databases_location)

    @property
    def secrets(self) -> ShellProfileSecrets:
        return ShellProfileSecrets.load(profile=self, location=self._databases_location)

    @property
    def events(self) -> ShellProfileEventsDatabase:
        return ShellProfileEventsDatabase.load(location=self._databases_location)

    @property
    def distro(self) -> Optional[Distro]:
        if self.settings.distro is None:
            return None
        if self.settings.distro not in KNOWN_DISTRIBUTIONS:
            logger.warning(f"Your profile is set to use the distribution '{self.settings.distro}' but this "
                           f"is not in the list of known distributions. Known distributions are "
                           f"{list(KNOWN_DISTRIBUTIONS.keys())}")
            return None
        return KNOWN_DISTRIBUTIONS[self.settings.distro]

    @distro.setter
    def distro(self, value: Union[str, Distro]):
        if isinstance(value, Distro):
            value = value.name
        assert value in KNOWN_DISTRIBUTIONS
        self.settings.distro = value

    def database(self, name: str, cls: Optional[Type[DTShellDatabase]] = None) -> DTShellDatabase:
        if cls is None:
            cls = DTShellDatabase
        return cls.open(name, location=self._databases_location)

    def needs_update(self, key: str, period: float, default: bool = True) -> bool:
        # read record
        try:
            last_time_checked: float = self.updates_check_db.get(key)
        except DTShellDatabase.NotFound:
            return default
        # check time
        return time.time() - last_time_checked > period

    def mark_updated(self, key: str, when: float = None):
        # update record
        self.updates_check_db.set(key, when if when is not None else time.time())

    def configure(self, readonly: bool = False) -> bool:
        modified_config: bool = False
        # make sure we have a distro for this profile
        if self.distro is None:
            if readonly:
                raise ConfigNotPresent()
            # see if we can find the distro ourselves
            matched: bool = False
            for d in KNOWN_DISTRIBUTIONS.keys():
                if d == self.name:
                    logger.info(f"Automatically selecting distribution '{d}' as it matches the profile name")
                    self.distro = d
                    matched = True
                    break

            # if we don't have a match, we ask the user to pick a distribution
            if not matched:
                print()
                print("You need to choose the distribution you want to work with in this profile.")
                distros: List[Choice] = []
                for distro in KNOWN_DISTRIBUTIONS.values():
                    # only show production branches
                    if distro.staging:
                        continue
                    # only show stable distributions
                    if not distro.stable:
                        continue
                    # ---
                    eol: str = "" if distro.end_of_life is None else \
                        f"(end of life: {distro.end_of_life_fmt})"
                    label = [("class:choice", distro.name), ("class:disabled", f"  {eol}")]
                    choice: Choice = Choice(title=label, value=distro.name)
                    if distro.name == SUGGESTED_DISTRIBUTION:
                        distros.insert(0, choice)
                    else:
                        distros.append(choice)
                # let the user choose the distro
                chosen_distro: str = questionary.select(
                    "Choose a distribution:", choices=distros, style=cli_style).unsafe_ask()
                # attach distro to profile
                self.distro = chosen_distro
            modified_config = True

        # make sure we have a token for this profile
        if self.secrets.dt_token is None:
            if readonly:
                raise ConfigNotPresent()
            print()
            print(f"The Duckietown Shell needs a Duckietown Token to work properly. "
                  f"Get yours for free at {DUCKIETOWN_TOKEN_URL}")
            while True:
                # let the user insert the token
                token_str: str = questionary.password("Enter your token:", validate=validator_token)\
                    .unsafe_ask()
                token: DuckietownToken = DuckietownToken.from_string(token_str)
                # make sure this token is supported by this profile distro
                tokens_supported: List[str] = self.distro.tokens_supported
                if token.version not in tokens_supported:
                    print(f"Token version '{token.version}' not supported by this profile's distro. "
                          f"Only versions supported are {tokens_supported}.")
                    continue
                else:
                    print(f"Token verified successfully. Your ID is: {yellow_bold(token.uid)}")
                    break
            # store token
            self.secrets.dt_token = token_str
            modified_config = True
        # ---
        return modified_config



import logging
from typing import Optional, Sequence

import termcolor

from .utils import dark_yellow

__all__ = ["dts_print", "setup_logging_color"]


def dts_print(msg: str, color: Optional[str] = None, attrs: Sequence[str] = ()) -> None:
    """
    Prints a message to the user.
    """
    msg = msg.strip(" ")  # remove space
    print("")  # always separate
    lines = msg.split("\n")
    prefix = "dts : "
    filler = "    : "
    # filler = ' ' * len(prefix)

    for i, line in enumerate(lines):
        f = prefix if i == 0 else filler
        on_color = None
        line = line.replace("\t", " " * 4)
        line = termcolor.colored(line, color, on_color, list(attrs))
        s = "%s %s" % (dark_yellow(f), line)
        print(s)


def add_coloring_to_emit_ansi(fn):
    # add methods we need to the class
    def new(*args):
        levelno = args[1].levelno
        if levelno >= 50:
            color = "\x1b[31m"  # red
        elif levelno >= 40:
            color = "\x1b[31m"  # red
        elif levelno >= 30:
            color = "\x1b[33m"  # yellow
        elif levelno >= 20:
            color = "\x1b[32m"  # green
        elif levelno >= 10:
            color = "\x1b[35m"  # pink
        else:
            color = "\x1b[0m"  # normal

        msg = str(args[1].msg)

        lines = msg.split("\n")

        def color_line(line):
            return "%s%s%s" % (color, line, "\x1b[0m")  # normal

        lines = list(map(color_line, lines))

        args[1].msg = "\n".join(lines)
        return fn(*args)

    return new


def setup_logging_color():
    import platform

    if platform.system() != "Windows":
        emit2 = add_coloring_to_emit_ansi(logging.StreamHandler.emit)
        logging.StreamHandler.emit = emit2



import argparse
from typing import Optional, List, Set

import questionary
from questionary import Choice

from dt_shell import DTCommandAbs, DTShell, UserAborted, UserError, dtslogger
from dt_shell.utils import cli_style


class DTCommand(DTCommandAbs):
    help = 'Switch to a different profile'

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # parse arguments
        parsed: argparse.Namespace = DTCommand.parser.parse_args(args)
        # get list of existing profiles
        profiles: Set[str] = set(shell.profiles.keys())
        # ask the user what profile to switch to if none is given
        new_profile: str
        if parsed.profile is not None:
            # make sure the given profile exists
            if parsed.profile not in profiles:
                raise UserError(f"Profile '{parsed.profile}' does not exist")
            new_profile = parsed.profile
        else:
            dtslogger.info("Choose the profile you want to switch to")
            distros: List[Choice] = []
            for profile in profiles:
                extras: str = "" if profile != shell.profile.name else f"(current)"
                label = [
                    ("class:choice", profile),
                    ("class:disabled", f" {extras}")
                ]
                choice: Choice = Choice(title=label, value=profile)
                distros.append(choice)
            # let the user choose the profile
            new_profile: Optional[str] = questionary.select(
                "Choose a profile:", choices=distros, style=cli_style).unsafe_ask()
            if new_profile is None:
                raise UserAborted()

        # same profile?
        if new_profile == shell.profile.name:
            dtslogger.info(f"Already on profile '{new_profile}'")
            return
        # switch profile
        dtslogger.info(f"Switching to profile '{new_profile}'...")
        # set the new profile as the profile to load at the next launch
        shell.settings.profile = new_profile
        dtslogger.info(f"Active profile is now set to '{new_profile}'")

    @staticmethod
    def complete(shell: DTShell, word: str, line: str) -> List[str]:
        return []



import argparse
import logging
from typing import Optional, List, Set

import questionary
from questionary import Choice

from dt_shell import DTCommandAbs, DTShell, UserAborted, UserError, dtslogger
from dt_shell.constants import KNOWN_DISTRIBUTIONS, SUGGESTED_DISTRIBUTION, EMBEDDED_COMMAND_SET_NAME
from dt_shell.profile import ShellProfile
from dt_shell.utils import cli_style
from dt_shell_cli import logger


class DTCommand(DTCommandAbs):
    help = 'Create new profile'

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # parse arguments
        parsed: argparse.Namespace = DTCommand.parser.parse_args(args)
        # get list of existing profiles
        profiles: Set[str] = set(shell.profiles.keys())
        # ask the user for a distribution to use
        dtslogger.info("In order to make a new profile, you need to choose a distribution first.")
        distros: List[Choice] = []
        for distro in KNOWN_DISTRIBUTIONS.values():
            # filter by staging VS production
            if parsed.staging != distro.staging:
                continue
            # only show stable distributions
            if not distro.stable and not parsed.unstable:
                continue
            # create questionnaire
            extras: str = "" if distro.name not in profiles else f"(profile already exists)"
            eol: str = "" if distro.end_of_life is None else f"(end of life: {distro.end_of_life_fmt})"
            label = [
                ("class:choice", distro.name),
                ("class:disabled", f"    {eol}"),
                ("class:disabled", f" {extras}")
            ]
            choice: Choice = Choice(title=label, value=distro.name)
            if distro.name.startswith(SUGGESTED_DISTRIBUTION) and distro.name not in profiles:
                distros.insert(0, choice)
            else:
                distros.append(choice)
        # let the user choose the distro
        new_profile: Optional[str] = questionary.select(
            "Choose a distribution:", choices=distros, style=cli_style).unsafe_ask()
        if new_profile is None:
            raise UserAborted()
        # let's make sure the user did not choose a distro for which a profile already exists
        if new_profile in profiles:
            raise UserError(f"A profile for the distribution '{new_profile}' already exists. "
                            f"Use the following command to switch profile."
                            f"\n\n\n\t\tdts profile switch {new_profile}\n\n")
        # make a new profile
        dtslogger.info(f"Setting up a new profile '{new_profile}'...")
        profile: ShellProfile = ShellProfile(name=new_profile, _distro=new_profile)
        # set the new profile as the profile to load at the next launch
        shell.settings.profile = new_profile
        # configure profile
        profile.configure()
        # download command sets
        logger.setLevel(logging.INFO)
        for cs in profile.command_sets:
            if cs.name == EMBEDDED_COMMAND_SET_NAME:
                continue
            if cs.leave_alone:
                continue
            # update command set
            cs.update()
        logger.setLevel(logging.WARNING)

    @staticmethod
    def complete(shell: DTShell, word: str, line: str) -> List[str]:
        return []



# -*- coding: utf-8 -*-

import argparse
import atexit
import dataclasses
import inspect
import os
import random
import signal
import sys
import time
import traceback
import types
from cmd import Cmd
from dataclasses import dataclass
from enum import Enum
from math import floor
from typing import List, Optional, Tuple, Union, Type, Dict, Callable
from typing import Mapping, Sequence

import questionary
from pyfiglet import Figlet
from questionary import Choice
from termcolor import termcolor

from . import __version__, logger, compatibility
from .checks.version import check_for_updates
from .commands import DTCommandAbs, CommandDescriptor, DTCommandPlaceholder, DTCommandConfigurationAbs, \
    CommandSet, NoOpCommand
from .commands.importer import import_command, import_configuration
from .compatibility.migrations import \
    migrate_distro, \
    needs_migrate_docker_credentials, migrate_docker_credentials, \
    needs_migrate_token_dt1, migrate_token_dt1, \
    needs_migrate_secrets, migrate_secrets, mark_docker_credentials_migrated, \
    mark_token_dt1_migrated, mark_secrets_migrated, needs_migrations, mark_all_migrated
from .constants import DNAME, KNOWN_DISTRIBUTIONS, SUGGESTED_DISTRIBUTION, EMBEDDED_COMMAND_SET_NAME, \
    DB_BILLBOARDS, DB_UPDATES_CHECK, CHECK_BILLBOARD_UPDATE_SECS, PUSH_USER_EVENTS_TO_HUB_SECS
from .constants import DTShellConstants, IGNORE_ENVIRONMENTS, DB_SETTINGS, DB_PROFILES
from .database import DTShellDatabase
from .environments import ShellCommandEnvironmentAbs, DEFAULT_COMMAND_ENVIRONMENT
from .exceptions import UserError, NotFound, CommandNotFound, CommandsLoadingException, UserAborted, \
    ConfigNotPresent
from .logging import dts_print
from .profile import ShellProfile
from .utils import text_justify, text_distribute, cli_style, indent_block, ensure_bash_completion_installed, \
    env_option

BILLBOARDS_VERSION: str = "v1"

CommandName = str
CommandsTree = Dict[CommandName, Union[None, Mapping[CommandName, dict], CommandDescriptor]]


@dataclass
class CLIOptions:
    debug: bool = env_option("DTSHELL_DEBUG", False)
    verbose: bool = env_option("DTSHELL_VERBOSE", False)
    quiet: bool = env_option("DTSHELL_QUIET", False)
    complete: bool = False
    profile: Optional[str] = env_option("DTSHELL_PROFILE", None)


def get_cli_options(args: List[str]) -> Tuple[CLIOptions, List[str]]:
    """Returns cli options plus other arguments for the commands."""
    default_opts: CLIOptions = CLIOptions()

    if args and not args[0].startswith("-"):
        return default_opts, args

    # find first non-option word
    i: int = 0
    for w in args:
        if w.startswith("-"):
            i += 1
            if w == "--profile":
                i += 1
        else:
            break

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--debug",
        action="store_true",
        default=default_opts.debug,
        help="More debug information"
    )
    parser.add_argument(
        "-vv", "--verbose",
        action="store_true",
        default=default_opts.verbose,
        help="More debug information from the shell"
    )
    parser.add_argument(
        "-q", "--quiet",
        action="store_true",
        default=default_opts.quiet,
        help="Quiet execution"
    )
    parser.add_argument(
        "--profile",
        type=str,
        default=default_opts.profile,
        help="Select specific profile just for this session"
    )

    if "--complete" in args[:i]:
        parser.add_argument(
            "--complete",
            action="store_true",
            default=False,
            help="Execute command completion",
        )

    parsed = parser.parse_args(args[:i])

    return CLIOptions(**parsed.__dict__), args[i:]


prompt = "dts> "


class ShellSettings(DTShellDatabase):

    @classmethod
    def load(cls, location: Optional[str] = None):
        return DTShellDatabase.open(DB_SETTINGS, location=location)

    @property
    def check_for_updates(self) -> bool:
        return self.get("check_for_updates", True)

    @check_for_updates.setter
    def check_for_updates(self, value: bool):
        assert isinstance(value, bool)
        self.set("check_for_updates", value)

    @property
    def profile(self) -> Optional[str]:
        return self.get("profile", None)

    @profile.setter
    def profile(self, value: str):
        assert isinstance(value, str)
        self.set("profile", value)

    @property
    def show_billboards(self) -> bool:
        return self.get("show_billboards", True)

    @show_billboards.setter
    def show_billboards(self, value: bool):
        assert isinstance(value, bool)
        self.set("show_billboards", value)


class EventType(Enum):
    START = "start"
    PRE_COMMAND_IMPORT = "pre-command-import"
    POST_COMMAND_IMPORT = "post-command-import"
    KEYBOARD_INTERRUPT = "keyboard-interrupt"
    SHUTDOWN = "shutdown"


@dataclasses.dataclass
class Event:
    type: EventType
    origin: str
    time: float = dataclasses.field(default_factory=time.time)


class DTShell(Cmd):
    commands: CommandsTree = {}
    core_commands: List[CommandName] = [
        "commands",
        "install",
        "uninstall",
        "update",
        "version",
        "exit",
        "help",
    ]

    # tree of commands once loaded
    include: types.SimpleNamespace

    def __init__(self,
                 skeleton: bool = False,
                 readonly: bool = False,
                 banner: bool = True,
                 billboard: bool = True,
                 profile: Optional[str] = None
                 ):
        # populate singleton
        import dt_shell
        dt_shell.shell = self

        # arguments
        self._skeleton: bool = skeleton
        self._readonly: bool = readonly
        self._banner: bool = banner
        self._billboard: bool = billboard

        # errors while loading end up in here
        self._errors_loading: List[str] = []

        # updates check database
        self.updates_check_db: DTShellDatabase[float] = DTShellDatabase.open(DB_UPDATES_CHECK)

        # namespace will contain the map to the loaded commands
        DTShell.include = types.SimpleNamespace()

        # event handlers
        self._event_handlers: Dict[EventType, List[Callable]] = {
            EventType.START: [
                self._run_background_tasks
            ],
            EventType.PRE_COMMAND_IMPORT: [],
            EventType.POST_COMMAND_IMPORT: [],
            EventType.KEYBOARD_INTERRUPT: [
                self._on_keyboard_interrupt_event
            ],
            EventType.SHUTDOWN: [
                self._on_shutdown_event
            ],
        }

        # set all databases to readonly if needed
        DTShellDatabase.global_readonly = readonly

        # open databases
        self._db_profiles: DTShellDatabase = DTShellDatabase.open(DB_PROFILES, readonly=readonly)
        self._db_settings: ShellSettings = ShellSettings.open(DB_SETTINGS, readonly=readonly)

        # custom profile
        if profile is not None:
            if profile not in self._db_profiles.keys():
                raise UserError(f"The profile '{profile}' does not exist.")
            logger.info(f"Using profile '{profile}' as prescribed by --profile or environment variable DTSHELL_PROFILE")
            with self.settings.in_memory():
                self.settings.profile = profile

        # load current profile
        self._profile: ShellProfile = ShellProfile(self.settings.profile, readonly=readonly) \
            if self.settings.profile else None

        # start event
        self._trigger_event(Event(EventType.START, "shell"))

        # get billboard to show (if any)
        bboard: Optional[str] = None
        if billboard and self.settings.show_billboards:
            bboard = self.get_billboard()

        # print banner
        if banner:
            self._show_banner(profile=self._profile, billboard=bboard)

        # make sure the bash completion script is installed
        if not readonly:
            ensure_bash_completion_installed()

        # check if we configure the shell by migrating an old profile
        self.performed_migrations: bool = self._attempt_migrations(readonly)

        # make sure the shell is configured
        self.configured_shell: bool = self._configure(readonly)

        # make sure the profile is configured
        self.configured_profile: bool = self._profile.configure(readonly)

        # in readonly mode we stop right here if we don't have a profile
        if readonly and self._profile is None:
            return

        # make sure the shell is properly configured
        assert self._profile is not None

        # update shell constants
        DTShellConstants.PROFILE = self._profile
        DTShellConstants.ROOT = self._profile.path

        # check for updates
        if not readonly and not skeleton and self.settings.check_for_updates:
            check_for_updates()

        # add command set path to PYTHONPATH
        for cs in self.command_sets:
            for path in [cs.path, os.path.join(cs.path, "lib")]:
                # add new path to PYTHONPATH for this session
                path: str = os.path.abspath(path)
                sys.path.insert(0, path)
                logger.debug(f"Path '{path}' added to PYTHONPATH")

        # add custom PYTHONPATH
        if "DTSHELL_PYTHONPATH" in os.environ:
            for path in os.environ["DTSHELL_PYTHONPATH"].split(":")[::-1]:
                # add new path to PYTHONPATH for this session
                path: str = os.path.abspath(path)
                if path in sys.path:
                    sys.path.remove(path)
                sys.path.insert(0, path)
                logger.info(f"Path '{path}' added to PYTHONPATH from DTSHELL_PYTHONPATH")

        # call super constructor
        super(DTShell, self).__init__()

        # remove the char `-` from the list of word separators, this allows us to suggest flags
        if self.use_rawinput and self.completekey:
            import readline
            readline.set_completer_delims(readline.get_completer_delims().replace("-", "", 1))

        # check for updates (if needed)
        if not readonly:
            for cs in self.command_sets:
                # Do not check it if we are using custom commands (leave-alone)
                if not cs.leave_alone:
                    cs.update()

        # pre-import event
        self._trigger_event(Event(EventType.PRE_COMMAND_IMPORT, "shell"))

        # load commands
        self.load_commands(skeleton)

        # make sure nobody is importing command implementations when in skeleton mode
        if skeleton:
            terminate: bool = False
            for subclass in DTCommandAbs.__subclasses__():
                if subclass in [DTCommandPlaceholder, NoOpCommand]:
                    continue
                origin_fpath: str = inspect.getfile(subclass)
                logger.error(f"The file '{origin_fpath}' was loaded while the shell run in skeleton "
                             f"mode. This is not allowed, command implementation should never be manually "
                             f"imported in the command set files.")
                terminate = True
            if terminate:
                raise UserError("Some command implementations were imported while running in skeleton mode.")

        # post-import event
        self._trigger_event(Event(EventType.POST_COMMAND_IMPORT, "shell"))

        # apply backward-compatibility edits
        compatibility.apply(self)

        # register SIGINT handler
        # TODO: disabled for now, we need to figure out how to handle this properly, commands naturally
        #  expect SIGINT signals and KeyboardInterrupt exceptions
        # signal.signal(
        #     signal.SIGINT,
        #     lambda sig, frame: self._trigger_event(Event(EventType.KEYBOARD_INTERRUPT, "user"))
        # )

        # register at-exit (we use a lambda so that the event is created at the proper time)
        atexit.register(lambda: self._trigger_event(Event(EventType.SHUTDOWN, "shell")))

    @property
    def profile(self) -> Optional[ShellProfile]:
        return self._profile

    @property
    def settings(self) -> ShellSettings:
        return self._db_settings

    @property
    def profiles(self) -> DTShellDatabase:
        return self._db_profiles

    @property
    def command_sets(self) -> List[CommandSet]:
        return self._profile.command_sets

    @property
    def commands_tree(self) -> List[CommandSet]:
        return self._profile.command_sets

    def command_set(self, name: str) -> CommandSet:
        """
        Returns the CommandSet with the given name if found installed. Raises NotFound otherwise.
        """
        for cs in self.command_sets:
            if cs.name == name:
                return cs
        raise NotFound(f"Command set '{name}' not found")

    def postcmd(self, stop, line):
        if len(line.strip()) > 0:
            print("")

    def default(self, line: str) -> None:
        # TODO: suggest possible commands as well
        dts_print(f"Unknown syntax:\n\n\t\tdts {line}\n", color="red")

    def emptyline(self):
        pass

    def complete(self, text, state):
        res = super(DTShell, self).complete(text, state)
        if res is not None:
            res += " "
        return res

    def on_event(self, event: EventType, handler: Callable[[Event], None]):
        if event not in self._event_handlers:
            raise KeyError(f"Event name '{event.name}' not recognized. "
                           f"Known events are: {list(self._event_handlers.keys())}")
        self._event_handlers[event].append(handler)

    def on_start(self, handler: Callable[[Event], None]):
        self.on_event(EventType.START, handler)

    def on_shutdown(self, handler: Callable[[Event], None]):
        self.on_event(EventType.SHUTDOWN, handler)

    def on_keyboard_interrupt(self, handler: Callable[[Event], None]):
        self.on_event(EventType.KEYBOARD_INTERRUPT, handler)

    def needs_update(self, key: str, period: float, default: bool = True) -> bool:
        # read record
        try:
            last_time_checked: float = self.updates_check_db.get(key)
        except DTShellDatabase.NotFound:
            return default
        # check time
        return time.time() - last_time_checked > period

    def is_time(self, key: str, period: float, default: bool = True) -> bool:
        return self.needs_update(key=key, period=period, default=default)

    def mark_updated(self, key: str, when: float = None):
        # update record
        self.updates_check_db.set(key, when if when is not None else time.time())

    def mark_done(self, key: str, when: float = None):
        self.mark_updated(key=key, when=when)

    def _run_background_tasks(self, event: Event):
        # we don't run background tasks in skeleton mode
        if self._readonly or self._skeleton:
            return
        if event.type is EventType.START:
            # update billboards
            if self.needs_update("billboards", CHECK_BILLBOARD_UPDATE_SECS):
                from .tasks import UpdateBillboardsTask
                UpdateBillboardsTask(self).start()
            # get docker versions
            from .tasks import CollectDockerVersionTask
            CollectDockerVersionTask(self).start()
            # push user events to the hub
            if self.is_time("upload_events", PUSH_USER_EVENTS_TO_HUB_SECS):
                from .tasks import UploadStatisticsTask
                UploadStatisticsTask(self).start()

    def _on_keyboard_interrupt_event(self, event: Event):
        pass

    def _on_shutdown_event(self, event: Event):
        pass

    def _trigger_event(self, event: Event):
        logger.debug(f"{event.origin.title()} triggered the event '{event.type.name}'")
        for cb in self._event_handlers[event.type]:
            try:
                cb(event)
            except Exception:
                traceback.print_exc()
                logger.error(f"An handler for the event '{event.type.name}' failed its execution. "
                             f"The exception is printed to screen.")

    def _attempt_migrations(self, readonly: bool = False) -> bool:
        modified_config: bool = False
        # make sure we need migrations
        if not needs_migrations():
            return modified_config
        elif readonly:
            raise ConfigNotPresent()

        asked_confirmation: bool = False
        # get the name of the old profile
        distro: Optional[str] = migrate_distro(dryrun=True)
        # if we didn't have a version set in the old format then we have nothing to migrate
        if distro is None:
            # we mark everything as migrated, so we don't ask again
            mark_all_migrated()
            return modified_config

        def _ask_confirmation() -> bool:
            print(f"The Duckietown shell now uses a new profile format. "
                  f"An old profile '{distro}' was found on disk. "
                  f"We can automatically migrate all your preferences, credentials and tokens into the "
                  f"new profile format.")
            # ask the user whether to continue with the migration
            return questionary.confirm(
                "Do you want to migrate your old profile?",
                auto_enter=True
            ).unsafe_ask()

        # try to migrate profile/distro
        if self.profile is None:
            granted: bool = _ask_confirmation()
            asked_confirmation = True
            if not granted:
                print("Nothing is better than a fresh start!")
                # we mark everything as migrated, so we don't ask again
                mark_all_migrated()
                return modified_config
            # we are migrating
            distro: str = migrate_distro(dryrun=True)
            # make new profile
            logger.info(f"Migrating profile '{distro}'...")
            self._profile = ShellProfile(name=distro)
            # set profile distro
            self._profile.distro = distro
            # set the new profile as the profile to load at the next launch
            self.settings.profile = distro
            modified_config = True

        # by now we must have a profile
        assert self.profile is not None

        # try to migrate dt1 token
        if "dt1" in self.profile.distro.tokens_supported and needs_migrate_token_dt1():
            migrate: bool = True
            if not asked_confirmation:
                migrate = _ask_confirmation()
                asked_confirmation = True
            # migrate?
            if migrate:
                token: Optional[str] = migrate_token_dt1(self.profile)
                if token is not None:
                    logger.info(f"Migrated: Tokens")
                # mark it as migrated, so we don't ask again
                mark_token_dt1_migrated()
            modified_config = True

        # try to migrate docker credentials
        if needs_migrate_docker_credentials():
            migrate: bool = True
            if not asked_confirmation:
                migrate = _ask_confirmation()
                asked_confirmation = True
            # migrate?
            if migrate:
                no_migrated: int = migrate_docker_credentials(self.profile)
                logger.info(f"Migrated: {no_migrated} Docker credentials")
                # mark it as migrated, so we don't ask again
                mark_docker_credentials_migrated()
            modified_config = True

        # try to migrate secrets
        if needs_migrate_secrets():
            migrate: bool = True
            if not asked_confirmation:
                migrate = _ask_confirmation()
                # noinspection PyUnusedLocal
                asked_confirmation = True
            # migrate?
            if migrate:
                migrate_secrets(self.profile)
                logger.info(f"Migrated: Other secrets")
                # mark it as migrated, so we don't ask again
                mark_secrets_migrated()
            modified_config = True

        # complete profile configuration
        modified_profile: bool = self.profile.configure()
        modified_config = modified_config or modified_profile
        return modified_config

    def load_commands(self, skeleton: bool):
        # rediscover commands
        self.commands = {}
        for cs in self.command_sets:
            # run command set init script
            if not skeleton:
                cs.init()

            # load commands from disk
            for cmd, subcmds in cs.commands.items():
                # noinspection PyTypeChecker
                self._load_command_subtree(cs, "", cmd, subcmds, 0, skeleton)

            # add commands to the list of commands
            self.commands.update(cs.commands)

        if len(self.commands) <= 0:
            logger.error("No commands found.")
            self.commands = {}

        if self._errors_loading:
            sep = "-" * 128
            msg = f"\n\n\n!   Could not load commands. Detailed error messages are printed below.\n" + \
                  indent_block(
                      f"\n\n{sep}\n\n\n" +
                      ("\n\n" + sep + "\n\n\n").join(self._errors_loading) +
                      f"\n\n{sep}\n\n"
                  ) + \
                  f"\n\n!   Could not load commands. Detailed error messages are printed above.\n\n"
            logger.error(msg)
            raise CommandsLoadingException("Some commands could not be loaded. Detailed error messages are "
                                           "reported above.")

    def reload_commands(self, skeleton: bool):
        # remove installed commands
        installed_commands = self.commands.keys()
        for command in installed_commands:
            for a in ["do_", "complete_", "help_"]:
                if hasattr(DTShell, a + command):
                    delattr(DTShell, a + command)
        # rediscover commands
        self.load_commands(skeleton)

    def _load_command_subtree(
        self,
        command_set: CommandSet,
        package: str,
        command: str,
        sub_commands: Union[None, Mapping[str, object], CommandDescriptor],
        lvl: int,
        skeleton: bool,
    ) -> Optional[Type[DTCommandAbs]]:
        # make a new (temporary) class
        class klass(DTCommandPlaceholder):
            pass
        selector: str = f"{package}{command}"
        configuration: Type[DTCommandConfigurationAbs] = import_configuration(command_set, selector)

        if isinstance(sub_commands, CommandDescriptor):
            descriptor: CommandDescriptor = sub_commands

            # load command configuration
            descriptor.configuration = configuration

            # if we ignore environments, we assume default global environment for every command
            environment: ShellCommandEnvironmentAbs
            if IGNORE_ENVIRONMENTS:
                environment = DEFAULT_COMMAND_ENVIRONMENT
            else:
                # figure out the environment for this command
                environment = descriptor.configuration.environment()
                if environment is None:
                    # revert to command set's default environment
                    environment = command_set.configuration.default_environment()
                if environment is None:
                    # use default environment
                    environment = DEFAULT_COMMAND_ENVIRONMENT

            # add environment to command's descriptor
            descriptor.environment = environment

            # import class only if this is the environment in which the commands will run
            if not skeleton:
                try:
                    klass = import_command(command_set, descriptor.path)
                except UserError:
                    raise
                except (UserAborted, KeyboardInterrupt):
                    raise
                except ModuleNotFoundError:
                    lines: List[str] = []
                    cs_path: str = os.path.abspath(os.path.realpath(command_set.path))
                    # check PYTHONPATH
                    found: bool = False
                    lines += ["\tPYTHONPATH: ["]
                    for p in sys.path:
                        p_path: str = os.path.abspath(os.path.realpath(p))
                        if p_path == cs_path:
                            found = True
                        lines += [f"\t\t'{p_path}'"]
                    lines += ["\t]"]
                    lines += [f"\t- dir[{cs_path}] in PYTHONPATH: {found}"]
                    # module already loaded?
                    m: str = ""
                    for p in selector.split("."):
                        m = f"{m}.{p}".lstrip(".")
                        mod = sys.modules.get(m, None)
                        if mod:
                            lines.append(f"\t- module[{m}] already loaded: True; {dir(mod)}")
                        else:
                            lines.append(f"\t- module[{m}] already loaded: False")

                    # check all __init__ files
                    fpath: str = os.path.join(cs_path)
                    for p in selector.split("."):
                        fpath = os.path.join(fpath, p)
                        init_fpath = os.path.join(fpath, "__init__.py")
                        lines.append(f"\t- file[{init_fpath}] exists: {os.path.isfile(init_fpath)}")
                    # compile details
                    details: str = "\n".join(lines)
                    msg = f"The command '{selector.replace('.', '/')}' could not be imported." \
                          f"\n\n{details}\n\n" \
                          f"{traceback.format_exc()}"
                    self._errors_loading.append(msg)
                except BaseException:
                    se = traceback.format_exc()
                    msg = (
                        f"Cannot load command class {descriptor.selector}.command.DTCommand "
                        f"(package={package}, command={command}):\n\n{se}"
                    )
                    self._errors_loading.append(msg)
                    return

            # link descriptor <-> command
            descriptor.command = klass
            klass.descriptor = descriptor
            # add loaded class to DTShell.include.<cmd_path>
            klass_path = [p for p in package.split(".") if len(p)]
            base = DTShell.include
            for p in klass_path:
                if not hasattr(base, p):
                    setattr(base, p, types.SimpleNamespace())
                base = getattr(base, p)
            setattr(base, command, klass)

        # give command its own info
        klass.name = command
        klass.level = lvl
        klass.parser = configuration.parser()
        # initialize list of subcommands
        klass.commands = {}

        # attach first-level commands to the shell
        if lvl == 0:
            do_command = getattr(klass, "do_command")
            get_command = getattr(klass, "get_command")
            complete_command = getattr(klass, "complete_command")
            help_command = getattr(klass, "help_command")
            # wrap [klass, function] around a lambda function
            do_command_lam = lambda s, w: do_command(s, w)
            get_command_lam = lambda s, w: get_command(s, w)
            complete_command_lam = lambda s, w, l, i, _: complete_command(s, w, l, i, _)
            help_command_lam = lambda s: help_command(s)
            # add functions do_* and complete_* to the shell
            for command_name in [command] + configuration.aliases():
                if DTShellConstants.VERBOSE:
                    logger.debug(f"Attaching root command '{command_name}' to shell")
                setattr(DTShell, "do_" + command_name, do_command_lam)
                setattr(DTShell, "get_" + command_name, get_command_lam)
                setattr(DTShell, "complete_" + command_name, complete_command_lam)
                setattr(DTShell, "help_" + command_name, help_command_lam)

        # stop recursion if there is no subcommand
        if sub_commands is None:
            return

        # load sub-commands
        if isinstance(sub_commands, dict):
            for cmd, subcmds in sub_commands.items():
                logger.debug("Searching %s at level %d" % (package + command + ".*", lvl))
                # noinspection PyTypeChecker
                kl = self._load_command_subtree(
                    command_set, package + command + ".", cmd, subcmds, lvl + 1, skeleton
                )
                if kl is not None:
                    klass.commands[cmd] = kl

        # return class for this command
        return klass

    def get_command(self, line) -> CommandDescriptor:
        """
        Interpret the argument and looks for the command that would be executed by the function onecmd(line).

        """
        cmd, arg, line = self.parseline(line)
        if not line or cmd is None or cmd == '':
            raise CommandNotFound(last_matched=None, remaining=line.split(" "))
        else:
            try:
                get_command = getattr(self, 'get_' + cmd)
            except AttributeError:
                raise CommandNotFound(last_matched=None, remaining=line.split(" "))
            # find command recursively down the tree
            cmd, _ = get_command(arg)
            return cmd

    # noinspection PyMethodMayBeStatic
    def sprint(self, msg: str, color: Optional[str] = None, attrs: Sequence[str] = None) -> None:
        attrs = attrs or []
        return dts_print(msg=msg, color=color, attrs=attrs)

    @staticmethod
    def get_billboard() -> Optional[str]:
        # get billboards from the local database
        db: DTShellDatabase = DTShellDatabase.open(DB_BILLBOARDS)
        # collect billboards names based on priority
        names: List[str] = []
        for name, billboard in db.items():
            names.extend([name] * (billboard["priority"] + 1))
        # no billboards?
        if not names:
            return None
        # pick one source at random
        name: str = random.choice(names)
        # ---
        return db.get(name).get("content", None)

    def update_commands(self):
        # update all command sets
        for cs in self.command_sets:
            if cs.name == EMBEDDED_COMMAND_SET_NAME:
                continue
            if cs.leave_alone:
                logger.warning(f"Will not update the command set '{cs.name}', it wants to be left alone.")
                continue
            # update command set
            logger.info(f"Updating the command set '{cs.name}'...")
            self.profile.events.new("shell/commandset/update", {"command_set": cs.as_dict()})
            cs.update()
            logger.info(f"Command set '{cs.name}' updated!")

    def _configure(self, readonly: bool = False) -> bool:
        modified_config: bool = False
        # make sure a profile exists and is loaded
        new_profile: Optional[str] = None
        if self._profile is None:
            if readonly:
                raise ConfigNotPresent()
            print()
            print("You need to choose the distribution you want to work with.")
            distros: List[Choice] = []
            for distro in KNOWN_DISTRIBUTIONS.values():
                # only show production branches
                if distro.staging:
                    continue
                # only show stable distributions
                if not distro.stable:
                    continue
                # ---
                eol: str = "" if distro.end_of_life is None else f"(end of life: {distro.end_of_life_fmt})"
                label = [("class:choice", distro.name), ("class:disabled", f"  {eol}")]
                choice: Choice = Choice(title=label, value=distro.name)
                if distro.name == SUGGESTED_DISTRIBUTION:
                    distros.insert(0, choice)
                else:
                    distros.append(choice)
            # let the user choose the distro
            new_profile = questionary.select(
                "Choose a distribution:", choices=distros, style=cli_style).unsafe_ask()
            modified_config = True

        # make a new profile if needed
        if new_profile is not None:
            if readonly:
                raise ConfigNotPresent()
            print(f"Setting up a new shell profile '{new_profile}'...")
            self._profile = ShellProfile(name=new_profile)
            # set the new profile as the profile to load at the next launch
            self.settings.profile = new_profile
            # configure profile
            modified_profile: bool = self._profile.configure()
            modified_config = modified_config or modified_profile
        # ---
        return modified_config

    def _show_banner(self, profile: Optional[ShellProfile], billboard: Optional[str]):
        width: int = 120
        if profile is None:
            # first launch -> bigger banner
            txt_width: int = 95  # measured
            padding: int = int(floor((width - txt_width) / 2))
            fmt: Figlet = Figlet(font='standard', width=width, justify='center')
            fig: str = fmt.renderText(DNAME.replace(" ", "   ").upper()).rstrip()
            sep: str = text_justify(termcolor.colored("_" * txt_width, "yellow", attrs=["bold"]), width)
            extras: str = text_distribute([
                f"First Setup - Welcome!",
                f"v{__version__}"
            ], width=txt_width)
            txt: str = f"{fig}\n{sep}\n{' ' * padding}{extras}\n\n"
        else:
            # any other launch -> smaller banner
            txt_width: int = 81  # measured
            padding: int = int(floor((width - txt_width) / 2))
            fmt: Figlet = Figlet(font='small', width=width, justify='center')
            fig: str = fmt.renderText(DNAME.replace(" ", "   ").upper()).rstrip()
            sep: str = text_justify(termcolor.colored("_" * txt_width, "yellow", attrs=["bold"]), width)
            extras: str = text_distribute([
                f"Profile: {self.settings.profile}",
                f"v{__version__}"
            ], width=txt_width)
            txt: str = f"{fig}\n{sep}\n{' ' * padding}{extras}\n\n"
        # ---
        print(txt.rstrip())
        print()
        print("+" + "-" * (width - 2) + "+")
        # print billboard (if given)
        if billboard:
            print("💬", billboard.strip())
            print("+" + "-" * (width - 2) + "+")




# -*- coding: utf-8 -*-
import json
import os
from datetime import datetime, timedelta
from typing import Optional, Tuple

import requests
import termcolor
import yaml
from requests import Response, HTTPError

from .. import __version__
from ..constants import DTShellConstants
from ..exceptions import CouldNotGetVersion, NoCacheAvailable, URLException


def get_url(url, timeout=3) -> str:
    try:
        response: Response = requests.get(url, timeout=timeout)
        response.raise_for_status()
        return response.text
    except HTTPError as e:
        raise URLException(str(e))
    except ConnectionError as e:
        raise URLException(str(e))


def get_last_version_fresh() -> str:
    url = "https://pypi.org/pypi/duckietown-shell/json"

    try:
        try:
            data = get_url(url)
        except URLException as e:
            raise CouldNotGetVersion(str(e))
        try:
            info = json.loads(data)
        except BaseException as e:
            msg = "Could not read json %r" % data
            raise CouldNotGetVersion(msg) from e

        last_version = info["info"]["version"]
        return last_version
    except CouldNotGetVersion:
        raise
    except BaseException as e:
        raise CouldNotGetVersion() from e


def get_cache_filename() -> str:
    d0 = os.path.expanduser(DTShellConstants.ROOT)
    return os.path.join(d0, "pypi-cache.yaml")


def read_cache() -> Tuple[str, datetime]:
    try:
        fn = get_cache_filename()
        if os.path.exists(fn):
            data = open(fn).read()
            interpreted = yaml.load(data, Loader=yaml.Loader)
            version = interpreted["version"]
            dt = interpreted["timestamp"]
            return version, dt
        else:
            msg = "File %s does not exist." % fn
            raise NoCacheAvailable(msg)
    except Exception as e:
        msg = "Could not read cache: %s" % e
        raise NoCacheAvailable(msg)


def write_cache(version: str, dt: datetime) -> None:
    fn = get_cache_filename()
    d0 = os.path.dirname(fn)
    if not os.path.exists(d0):
        os.makedirs(d0)

    with open(fn, "w") as f:
        contents = dict(version=version, timestamp=dt)
        y = yaml.dump(contents)
        f.write(y)


def get_last_version() -> Optional[str]:
    now = datetime.now()
    update = False

    version = None
    try:
        version, timestamp = read_cache()
    except NoCacheAvailable:
        timestamp = now
        update = True

    if not update:
        delta = now - timestamp
        if delta > timedelta(minutes=10):
            # logger.debug('Version cache is outdated (%s).' % delta)
            update = True

    if update:
        # logger.debug('Getting last version from PyPI.')
        try:
            version = get_last_version_fresh()
            write_cache(version, now)
            return version
        except CouldNotGetVersion:
            return None

    # XXX: this might not be set
    return version


def is_older(a: str, b: str) -> bool:
    na = tuple(map(int, a.split(".")))
    nb = tuple(map(int, b.split(".")))

    return na < nb


def check_for_updates() -> None:
    latest_version = get_last_version()
    # print('last version: %r' % latest_version)
    # print('installed: %r' % __version__)

    # TODO: this should take into account the profile we are using and the max major version declared in them

    if latest_version and is_older(__version__, latest_version):
        msg = """

There is an updated duckietown-shell available.

  You have: {current}

  Available: {available} 
 
WARNING: We strongly recommend updating to the latest version. ONLY THE LATEST VERSION IS SUPPORTED!
         If you experience issues, please make sure you're using the latest version before posting 
         questions or issues. 

You can update the shell using `pip`. Run the following command:
        pip3 install --no-cache-dir --user -U duckietown-shell

        """.format(
            current=__version__, available=latest_version
        )
        print(termcolor.colored(msg, "yellow"))



import copy
import os.path
from abc import abstractmethod, ABC
from contextlib import contextmanager
from threading import Semaphore
from typing import Union, TypeVar, Generic, Tuple, Optional, Dict, Iterator, ContextManager

import yaml
from filelock import FileLock, Timeout

from ..utils import safe_pathname

SerializedValue = Union[int, float, str, bytes, dict, list]
T = TypeVar('T', bound=SerializedValue)
Key = Union[str, Tuple[str]]


DEFAULT_DATABASES_DIR = os.path.expanduser("~/.duckietown/shell/databases/")
DATABASES_DIR = os.environ.get("DTSHELL_DATABASES", DEFAULT_DATABASES_DIR)

EMPTY_DB = {
    "version": 1,
    "data": {}
}


class DTSerializable(ABC):

    @abstractmethod
    def dump(self) -> SerializedValue:
        pass

    @classmethod
    @abstractmethod
    def load(cls, v: SerializedValue) -> 'DTSerializable':
        pass


NOTSET = object()
NATURALLY_SERIALIZABLE = (int, float, str, bytes, dict, list)
SerializableValue = Union[SerializedValue, DTSerializable]
SerializableTypes = (SerializedValue, *NATURALLY_SERIALIZABLE)


class DTShellDatabase(Generic[T]):

    _instances: Dict[Tuple[str, str], 'DTShellDatabase'] = {}
    global_readonly: bool = False

    class NotFound(KeyError):
        pass

    def __init__(self):
        self._name: str = ""
        self._location: str = ""
        self._readonly: bool = False
        self._data: dict = {}
        self._ephemeral: dict = {}
        self._lock: Semaphore = Semaphore()
        self._atomic: FileLock = FileLock(f"{self.yaml}.lock", timeout=10)
        self._in_memory: bool = False
        # ---
        raise RuntimeError(f'Call {self.__class__.__name__}.open() instead')

    @classmethod
    def open(cls, name: str, location: Optional[str] = DATABASES_DIR, readonly: bool = False,
             init_args: dict = None) -> 'DTShellDatabase':
        key = (location, name)
        if key not in cls._instances:
            # noinspection PyArgumentList
            inst = cls.__new__(cls, **(init_args or {}))
            cls._instances[key] = inst
            # populate instance fields
            inst._name = name
            inst._location = location
            inst._readonly = readonly or cls.global_readonly
            inst._data = {}
            inst._ephemeral = {}
            inst._lock = Semaphore()
            inst._atomic = FileLock(f"{inst.yaml}.lock", timeout=10)
            inst._in_memory = False
            # set custom init args
            for k, v in (init_args or {}).items():
                setattr(inst, k, v)
            # load DB from disk
            inst._load()
        # ---
        return cls._instances[key]

    @property
    def name(self) -> str:
        return self._name

    @property
    def yaml(self) -> str:
        yaml_fpath: str = os.path.abspath(os.path.join(self._location, f"{safe_pathname(self._name)}.yaml"))
        # make destination if it does not exist
        if not self._readonly:
            self._ensure_dir(yaml_fpath)
        # ---
        return yaml_fpath

    @contextmanager
    def in_memory(self) -> ContextManager:
        # code to acquire resource
        self._in_memory = True
        try:
            yield self
        finally:
            # code to release resource
            self._in_memory = False

    @property
    def size(self) -> int:
        return len(self._data)

    def contains(self, key: Key) -> bool:
        key = self._key(key)
        return key in self._data

    def get(self, key: Key, default: Optional[T] = NOTSET) -> T:
        key = self._key(key)
        # in memory?
        if key in self._ephemeral:
            return self._ephemeral[key]
        # read from persistent data
        try:
            return self._data[key]
        except KeyError:
            if default is NOTSET:
                raise DTShellDatabase.NotFound(f"Key '{key}' not found in database.")
            else:
                return default

    def delete(self, key: Key):
        key = self._key(key)
        # in memory?
        self._ephemeral.pop(key, None)
        # persistent data
        with self._lock:
            self._data.pop(key, None)
        self._write()

    def set(self, key: Key, value: T):
        key = self._key(key)
        with self._lock:
            # in memory?
            if self._in_memory:
                self._ephemeral[key] = value
            else:
                self._data[key] = value
                self._ephemeral.pop(key, None)
            # ---
        self._write()

    def keys(self) -> Iterator[Key]:
        with self._lock:
            data: dict = copy.copy(self._data)
        return iter(data.keys())

    def values(self) -> Iterator[T]:
        with self._lock:
            data: dict = copy.copy(self._data)
        return iter(data.values())

    def items(self) -> Iterator[Tuple[Key, T]]:
        with self._lock:
            data: dict = copy.copy(self._data)
        return iter(data.items())

    def clear(self):
        """
        Removes all the records from the database.
        """
        self._data.clear()
        self._write()

    def update(self, d: dict):
        """
        Update this database with the records from the given database.
        """
        self._data.update(d)
        self._write()

    def _load(self):
        if not self._readonly:
            # make files if they don't exist
            if not os.path.exists(self.yaml):
                self._write()
        # check if the file exists
        exists: bool = os.path.exists(self.yaml) and os.path.isfile(self.yaml)
        # read from disk
        if exists:
            try:
                with self._atomic:
                    with open(self.yaml, "rt") as fin:
                        content = yaml.safe_load(fin)
            except Timeout:
                raise TimeoutError(f"Could not acquire lock for '{self.yaml}'. "
                                   f"If this happens often, delete the file {self.yaml}.lock")
            # populate internal state
            self._data = content["data"]

    def _write(self):
        # skip writing to disk if in read-only mode
        if self._readonly:
            return
        # complete data with other metadata
        with self._lock:
            content = {**EMPTY_DB, "data": {**self._data}}
            # write to disk
            try:
                with self._atomic:
                    with open(self.yaml, "wt") as fout:
                        yaml.safe_dump(content, fout, indent=4)
            except Timeout:
                raise TimeoutError(f"Could not acquire lock for '{self.yaml}'. "
                                   f"If this happens often, delete the file {self.yaml}.lock")

    @staticmethod
    def _serialize(v: object) -> SerializedValue:
        if not isinstance(v, SerializableTypes):
            raise ValueError(f"Object of type '{str(type(v))}' is not serializable. Supported types are: "
                             f"{[c.__name__ for c in SerializableTypes]}")
        if isinstance(v, NATURALLY_SERIALIZABLE):
            return v
        if isinstance(v, DTSerializable):
            return v.dump()

    @staticmethod
    def _ensure_dir(f: str):
        os.makedirs(os.path.dirname(f), exist_ok=True)

    @staticmethod
    def _key(k: Key) -> str:
        return k



import argparse
import os
from typing import Optional

from argcomplete import CompletionFinder, split_line


class ArgumentParserCompleter(CompletionFinder):

    def __init__(self, argument_parser: Optional[argparse.ArgumentParser] = None):
        super(ArgumentParserCompleter, self).__init__(argument_parser=argument_parser, append_space=True)

    def get_completions(self, line: str):
        """
        This method is a wrapper around the method _get_completions from the superclass
        """
        # reproduce the traditional BASH setup for autocomplete with COMP_LINE and COMP_POINT
        comp_line: str = line
        comp_point: int = os.environ.get("COMP_POINT", len(line))

        # NOTE: DEBUG only
        # print(
        #     "\nLINE: {!r}".format(comp_line),
        #     "\nPOINT: {!r}".format(comp_point),
        # )

        cword_prequote, cword_prefix, cword_suffix, comp_words, last_wordbreak_pos = \
            split_line(comp_line, comp_point)
        comp_words = comp_words[1:]

        if cword_prefix and cword_prefix[0] in self._parser.prefix_chars and "=" in cword_prefix:
            # special case for when the current word is "--optional=PARTIAL_VALUE". Give optional to parser.
            comp_words.append(cword_prefix.split("=", 1)[0])

        # NOTE: DEBUG only
        # print(
        #     "\nLINE: {!r}".format(comp_line),
        #     "\nPOINT: {!r}".format(comp_point),
        #     "\nPREQUOTE: {!r}".format(cword_prequote),
        #     "\nPREFIX: {!r}".format(cword_prefix),
        #     "\nSUFFIX: {!r}".format(cword_suffix),
        #     "\nWORDS:", comp_words,
        # )

        return self._get_completions(comp_words, cword_prefix, cword_prequote, last_wordbreak_pos)



from .commands import \
    DTCommandAbs, \
    DTCommandPlaceholder, \
    DTCommandSetConfigurationAbs, \
    DTCommandConfigurationAbs, \
    CommandSet, \
    CommandDescriptor, \
    NoOpCommand, \
    FailedToLoadCommand, \
    noop_command, \
    failed_to_load_command, \
    default_command_configuration



from typing import cast, List, Optional

from ..exceptions import UserError
from ..utils import parse_version

__all__ = ["check_package_version", "_get_installed_distributions"]


def _get_installed_distributions(
    local_only: bool = True,
    user_only: bool = False,
    paths: Optional[List[str]] = None,
):
    """Return a list of installed Distribution objects."""
    from pip._internal.metadata import get_environment
    from pip._internal.metadata.pkg_resources import Distribution as _Dist

    env = get_environment(paths)
    dists = env.iter_installed_distributions(
        local_only=local_only,
        user_only=user_only,
        skip=[],
    )
    return [cast(_Dist, dist)._dist for dist in dists]


def check_package_version(PKG: str, min_version: str):
    # pip_version = "?"
    # try:
    #     # noinspection PyCompatibility
    #     from pip import __version__
    #
    #     pip_version = __version__
    #     # noinspection PyCompatibility
    #     from pip._internal.utils.misc import get_installed_distributions
    # except ImportError:
    #     msg = f"""
    #        You need a higher version of "pip".  You have {pip_version}
    #
    #        You can install it with a command like:
    #
    #            pip install -U pip
    #
    #        (Note: your configuration might require a different command.)
    #        """
    #     raise UserError(msg)

    installed = _get_installed_distributions()
    pkgs = {_.project_name: _ for _ in installed}
    if PKG not in pkgs:
        msg = f"""
        You need to have an extra package installed called `{PKG}`.

        You can install it with a command like:

            pip3 install -U "{PKG}>={min_version}"

        (Note: your configuration might require a different command.
         You might need to use "pip" instead of "pip3".)
        """
        raise UserError(msg)

    p = pkgs[PKG]

    installed_version = parse_version(p.version)
    required_version = parse_version(min_version)
    if installed_version < required_version:
        msg = f"""
       You need to have installed {PKG} of at least {min_version}.
       We have detected you have {p.version}.

       Please update {PKG} using pip.

           pip3 install -U  "{PKG}>={min_version}"

       (Note: your configuration might require a different command.
        You might need to use "pip" instead of "pip3".)
       """
        raise UserError(msg)



import os.path
import shutil
from typing import Optional, List

import yaml

from dt_shell.constants import DB_MIGRATIONS, KNOWN_DISTRIBUTIONS
from dt_shell.database import DTShellDatabase
from dt_shell.profile import ShellProfile, DockerCredentials, GenericCredentials, ShellProfileSecrets

OLD_ROOT: str = os.path.expanduser("~/.dt-shell")


def read_old_config() -> Optional[dict]:
    # old config file
    config_fpath: str = os.path.join(OLD_ROOT, "config.yaml")
    if not os.path.exists(config_fpath):
        return None
    with open(config_fpath, "rt") as fin:
        return yaml.safe_load(fin.read())


def needs_migrate_distro() -> bool:
    # check migrations database
    return not _already_migrated("distro")


def needs_migrate_docker_credentials() -> bool:
    # check migrations database
    return not _already_migrated("docker_credentials")


def needs_migrate_token_dt1() -> bool:
    # check migrations database
    return not _already_migrated("token_dt1")


def needs_migrate_secrets() -> bool:
    # check migrations database
    return not _already_migrated("secrets")


def needs_migrations() -> bool:
    return needs_migrate_docker_credentials() or \
           needs_migrate_token_dt1() or \
           needs_migrate_secrets()


def mark_distro_migrated():
    _mark_migrated("distro")


def mark_docker_credentials_migrated():
    _mark_migrated("docker_credentials")


def mark_token_dt1_migrated():
    _mark_migrated("token_dt1")


def mark_secrets_migrated():
    _mark_migrated("secrets")


def mark_all_migrated():
    mark_distro_migrated()
    mark_docker_credentials_migrated()
    mark_token_dt1_migrated()
    mark_secrets_migrated()


def migrate_docker_credentials(profile) -> int:
    profile: ShellProfile
    try:
        # old config file
        config: dict = read_old_config()
        if not config:
            return 0
        # open new database
        db: DockerCredentials = profile.secrets.docker_credentials
        # migrate credentials
        docker_credentials: dict = config.get("docker_credentials", {})
        for registry, credentials in docker_credentials.items():
            db.set(
                registry,
                GenericCredentials(username=credentials["username"], password=credentials["secret"])
            )
        # return the number of credentials migrated
        return len(docker_credentials)
    finally:
        # mark as migrated
        mark_docker_credentials_migrated()


def migrate_distro(dryrun: bool = False) -> Optional[str]:
    profile: ShellProfile
    try:
        # old config file
        config: dict = read_old_config()
        if not config:
            return None
        # get distro
        distro: str = config.get("duckietown_version", None)
        known_distros: List[str] = [d.name for d in KNOWN_DISTRIBUTIONS.values()]
        if distro not in known_distros:
            return None
        # ---
        return distro
    finally:
        # mark it as migrated
        if not dryrun:
            mark_distro_migrated()


def migrate_token_dt1(profile) -> Optional[str]:
    profile: ShellProfile
    try:
        # old config file
        config: dict = read_old_config()
        if not config:
            return None
        # open new database
        db: ShellProfileSecrets = profile.secrets
        # migrate credentials
        token_dt1: str = config.get("token_dt1", None)
        if not token_dt1:
            return None
        db.dt1_token = token_dt1
        # ---
        return token_dt1
    finally:
        # mark as migrated
        mark_token_dt1_migrated()


def migrate_secrets(profile):
    profile: ShellProfile
    try:
        # secrets dirs
        old_secrets_dir: str = os.path.join(OLD_ROOT, "secrets")
        new_secrets_dir: str = os.path.join(profile.path, "secrets")
        # skip if the old profile does not have a secrets dir
        if not os.path.exists(old_secrets_dir):
            return
        # copy all files
        _copy_and_overwrite(old_secrets_dir, new_secrets_dir)
    finally:
        # mark it as migrated
        mark_secrets_migrated()


def _copy_and_overwrite(from_path: str, to_path: str):
    shutil.copytree(from_path, to_path, dirs_exist_ok=True)


def _mark_migrated(key: str):
    migrations_db: DTShellDatabase[bool] = DTShellDatabase.open(DB_MIGRATIONS)
    migrations_db.set(key, True)


def _already_migrated(key: str) -> bool:
    migrations_db: DTShellDatabase[bool] = DTShellDatabase.open(DB_MIGRATIONS)
    return migrations_db.get(key, False)



from typing import List

from dt_shell import DTCommandAbs, DTShell


class DTCommand(DTCommandAbs):
    help = "Uninstalls a command set."

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # TODO: here we uninstall a command set
        return True

    @staticmethod
    def complete(shell: DTShell, word: str, line: str) -> List[str]:
        return []



from typing import List

from dt_shell import DTCommandAbs, DTShell


class DTCommand(DTCommandAbs):

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        print("Bye bye!")
        exit()



import logging
import warnings

warnings.filterwarnings(action='ignore', module='.*paramiko.*')

logging.basicConfig()

# logger dedicated to the shell
logger = logging.getLogger("shell")
logger.setLevel(logging.INFO)



import argparse
from typing import Optional, List

from dt_shell.commands import DTCommandConfigurationAbs


class DTCommandConfiguration(DTCommandConfigurationAbs):

    @classmethod
    def parser(cls, **kwargs) -> Optional[argparse.ArgumentParser]:
        pass

    @classmethod
    def aliases(cls) -> List[str]:
        return []



from typing import Optional, Tuple

import dt_shell

from dt_shell.commands import DTCommandSetConfigurationAbs
from dt_shell.environments import ShellCommandEnvironmentAbs, VirtualPython3Environment


class DTCommandSetConfiguration(DTCommandSetConfigurationAbs):

    @classmethod
    def default_environment(cls, *args, **kwargs) -> Optional[ShellCommandEnvironmentAbs]:
        """
        The environment in which the commands in this set will run.
        """
        return VirtualPython3Environment()

    @classmethod
    def version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        """
        Version of this command set in the format (major, minor, patch).
        """
        # noinspection PyTypeChecker
        return tuple(map(int, dt_shell.__version__.split(".")))

    @classmethod
    def minimum_shell_version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        """
        Minimum version of the shell supported in the format (major, minor, patch).
        """
        return 6, 0, 0

    @classmethod
    def maximum_shell_version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        """
        Maximum version of the shell supported in the format (major, minor, patch).
        """
        return 99, 0, 0



import logging
from typing import List

from dt_shell import DTCommandAbs


__all__ = ["DTCommand"]

from dt_shell import DTShell, logger


class DTCommand(DTCommandAbs):

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # set the internal logger to print INFO messages
        logger.setLevel(logging.INFO)
        # update commands
        shell.update_commands()



# used by the following commands files:
#
#   - update/command.py
#
import os
from types import SimpleNamespace

from .config import ShellConfig, read_shell_config


class OtherVersions:

    @classmethod
    @property
    def name2versions(cls):
        from dt_shell.utils import DebugInfo
        return DebugInfo.name2versions


def apply(shell):
    """
    Applies backward compatibility edits to an instance of the shell
    """

    def get_dt1_token() -> str:
        var = "token_dt1"
        from_env = os.environ.get(var, None)
        if from_env:
            msg = f"Using token from environment variable {var} instead of profile secrets."
            print(msg)
            return from_env
        if shell.profile.secrets.dt1_token is None:
            msg = 'Please set up a token for this using "dts tok set".'
            raise Exception(msg)
        else:
            return shell.profile.secrets.dt1_token

    def get_commands_version() -> str:
        return shell.profile.name

    def save_config():
        from dt_shell.profile import GenericCredentials
        for registry, credentials in shell.shell_config.docker_credentials.items():
            cred: GenericCredentials = GenericCredentials(
                username=credentials["username"],
                password=credentials["secret"],
            )
            shell.profile.secrets.docker_credentials.set(registry, cred)

    # add methods to the given shell instance
    shell.get_dt1_token = get_dt1_token
    shell.get_commands_version = get_commands_version
    shell.shell_config = read_shell_config(shell)
    shell.save_config = save_config
    shell.local_commands_info = SimpleNamespace(leave_alone=False)



import json
import traceback
from dataclasses import dataclass
from typing import Optional, List

from dt_shell_cli import logger
from ..exceptions import RunCommandException
from ..checks.version import get_url
from ..constants import DEFAULT_COMMAND_SET_REPOSITORY, DTShellConstants
from ..utils import run_cmd, provider_username_project_from_git_url, indent_block


@dataclass
class CommandsRepository:
    username: str
    project: str
    branch: str
    location: Optional[str] = None

    @property
    def remoteurl(self) -> str:
        return f"https://github.com/{self.username}/{self.project}"

    def as_dict(self) -> dict:
        return {
            "username": self.username,
            "project": self.project,
            "branch": self.branch,
            "location": self.location,
            "url": self.remoteurl,
        }

    def remote_sha(self) -> Optional[str]:
        # Get the remote sha from GitHub
        logger.info("Fetching remote SHA from github.com ...")
        remote_url: str = "https://api.github.com/repos/%s/%s/branches/%s" % (
            self.username,
            self.project,
            self.branch,
        )
        # contact github
        try:
            content = get_url(remote_url)
        except Exception:
            if DTShellConstants.VERBOSE:
                traceback.print_exc()
            logger.debug(f"URL called: {remote_url}")
            logger.warning(f"An error occurred while fetching the remote SHA for repository {self.remoteurl}")
            return None
        # parse output
        try:
            data = json.loads(content)
            return data["commit"]["sha"]
        except Exception:
            if DTShellConstants.VERBOSE:
                traceback.print_exc()
            logger.debug(f"URL called: {remote_url}\n"
                         f"Object returned by github:\n\n{indent_block(content)}")
            logger.warning(f"An error occurred while fetching the remote SHA for repository {self.remoteurl}")
            return None

    @classmethod
    def given_distro(cls, distro: str) -> 'CommandsRepository':
        return CommandsRepository(
            **DEFAULT_COMMAND_SET_REPOSITORY,
            branch=distro
        )

    @classmethod
    def from_file_system(cls, path: str, location: Optional[str] = None) -> 'CommandsRepository':
        origin_url = run_cmd(["git", "-C", path, "config", "--get", "remote.origin.url"])
        _, username, project = provider_username_project_from_git_url(origin_url)
        branch = run_cmd(["git", "-C", path, "rev-parse", "--abbrev-ref", "HEAD"])
        return CommandsRepository(username, project, branch, location=location)

    @staticmethod
    def head_tag(path: str) -> Optional[str]:
        try:
            tags: str = run_cmd(["git", "-C", path, "describe", "--exact-match", "--tags", "HEAD"])
        except RunCommandException as e:
            if e.stderr is not None and "no tag exactly matches" in e.stderr:
                # just not an exact match
                return None
            if DTShellConstants.VERBOSE:
                traceback.print_exc()
            return None
        except Exception:
            if DTShellConstants.VERBOSE:
                traceback.print_exc()
            return None
        # ---
        tags: List[str] = tags.strip().split("\n")
        if not tags:
            return None
        return tags[0]

    @staticmethod
    def closest_tag(path: str) -> Optional[str]:
        try:
            tags: str = run_cmd(["git", "-C", path, "tag", "--merged"])
        except Exception:
            if DTShellConstants.VERBOSE:
                traceback.print_exc()
            return None
        # ---
        tags: List[str] = tags.strip().split("\n")
        if not tags:
            return None
        return tags[-1]



import dataclasses
import os
import subprocess
import sys
import venv
from abc import ABCMeta, abstractmethod
from traceback import format_exc
from typing import Optional, List, Dict


from . import logger
from .exceptions import ShellInitException, InvalidEnvironment, CommandsLoadingException, UserError, \
    UserAborted
from .constants import SHELL_LIB_DIR, SHELL_REQUIREMENTS_LIST, DTShellConstants
from .database.utils import InstalledDependenciesDatabase
from .logging import dts_print
from .utils import pip_install, replace_spaces, print_debug_info


class ShellCommandEnvironmentAbs(metaclass=ABCMeta):

    @abstractmethod
    def execute(self, shell, args: List[str]):
        raise NotImplementedError("Subclasses should implement the function execute()")


@dataclasses.dataclass
class Python3Environment(ShellCommandEnvironmentAbs):
    """
    Python3 environment shared with the shell library.
    Default for all the distros up to and including 'daffy'.
    """

    def execute(self, shell, args: List[str]):
        from .shell import DTShell
        from dtproject.exceptions import DTProjectNotFound
        shell: DTShell
        # run shell
        known_exceptions = (InvalidEnvironment, CommandsLoadingException, DTProjectNotFound)
        try:
            args = map(replace_spaces, args)
            cmdline = " ".join(args)
            shell.onecmd(cmdline)
        except UserError as e:
            msg = str(e)
            dts_print(msg, "red")
            print_debug_info()
            sys.exit(1)
        except known_exceptions as e:
            msg = str(e)
            dts_print(msg, "red")
            print_debug_info()
            sys.exit(1)
        except SystemExit:
            raise
        except (UserAborted, KeyboardInterrupt):
            dts_print("User aborted operation.")
            pass
        except BaseException:
            msg = format_exc()
            dts_print(msg, "red", attrs=["bold"])
            print_debug_info()
            sys.exit(2)


@dataclasses.dataclass
class VirtualPython3Environment(ShellCommandEnvironmentAbs):
    """
    Virtual Python3 environment dedicated to a profile and NOT SHARED with the shell library.
    Default for the 'ente' distribution.
    """

    def execute(self, shell, _: List[str]):
        from .shell import DTShell
        shell: DTShell
        # ---
        # we make a virtual environment
        DTSHELL_VENV_DIR: str = os.environ.get("DTSHELL_VENV_DIR", None)
        venv_leave_alone: bool = False
        if DTSHELL_VENV_DIR:
            logger.info(
                f"Using virtual environment from '{DTSHELL_VENV_DIR}' as instructed by the environment "
                f"variable DTSHELL_VENV_DIR.")
            venv_dir: str = DTSHELL_VENV_DIR
            venv_leave_alone = True
        else:
            venv_dir: str = os.path.join(shell.profile.path, "venv")

        # define path to virtual env's interpreter
        interpreter_fpath: str = os.path.join(venv_dir, "bin", "python3")

        # make and configure env path if it does not exist
        # TODO: this is a place where a --hard-reset flag would ignore the fact that the venv already exists
        #  and make a new one
        if not os.path.exists(interpreter_fpath):
            if venv_leave_alone:
                msg: str = f"The custom Virtual Environment path '{venv_dir}' was given but no virtual " \
                           f"environments were found at that location."
                logger.error(msg)
                raise ShellInitException(msg)

            # make venv if it does not exist
            logger.info(f"Creating new virtual environment in '{venv_dir}'...")
            os.makedirs(venv_dir, exist_ok=True)
            venv.create(
                venv_dir,
                system_site_packages=False,
                clear=False,
                symlinks=True,
                with_pip=False,
                prompt="dts"
            )
            # install pip
            get_pip_fpath: str = os.path.join(SHELL_LIB_DIR, "assets", "get-pip.py")
            assert os.path.exists(get_pip_fpath)
            logger.info(f"Installing pip...")
            try:
                subprocess.check_output([interpreter_fpath, get_pip_fpath], stderr=subprocess.PIPE)
            except subprocess.CalledProcessError as e:
                # TODO: test this failure case on purpose
                msg: str = "An error occurred while installing pip in the virtual environment"
                raise ShellInitException(msg, stdout=e.stdout, stderr=e.stderr)

        # install dependencies
        cache: InstalledDependenciesDatabase = InstalledDependenciesDatabase.load(shell.profile)
        # - shell
        if DTShellConstants.VERBOSE:
            logger.debug("Checking for changes in the shell's dependencies list...")
        if cache.needs_install_step(SHELL_REQUIREMENTS_LIST):
            # warn user of detected changes (if any)
            if cache.contains(SHELL_REQUIREMENTS_LIST):
                logger.info("Detected changes in the dependencies list for the shell")
            # proceed with installing new dependencies
            logger.info("Installing shell dependencies...")
            pip_install(interpreter_fpath, SHELL_REQUIREMENTS_LIST)
            cache.mark_as_installed(SHELL_REQUIREMENTS_LIST)
        else:
            if DTShellConstants.VERBOSE:
                logger.debug("No new dependencies or constraints detected")
        # - command sets
        for cs in shell.command_sets:
            if DTShellConstants.VERBOSE:
                logger.debug(f"Checking for changes in the dependencies list for command set '{cs.name}'...")
            requirements_list: Optional[str] = cs.configuration.requirements()
            if cache.needs_install_step(requirements_list):
                # warn user of detected changes (if any)
                if cache.contains(requirements_list):
                    logger.info(f"Detected changes in the dependencies list for the command set '{cs.name}'")
                # proceed with installing new dependencies
                logger.info(f"Installing dependencies for command set '{cs.name}'...")
                pip_install(interpreter_fpath, requirements_list)
                cache.mark_as_installed(requirements_list)
            else:
                if DTShellConstants.VERBOSE:
                    logger.debug("No new dependencies or constraints detected")

        # run shell in virtual environment
        import dt_shell_cli
        main_py: str = os.path.join(os.path.abspath(dt_shell_cli.__path__[0]), "main.py")
        exec_args: List[str] = [interpreter_fpath, interpreter_fpath, main_py, *sys.argv[1:]]

        exec_env: Dict[str, str] = {
            **os.environ,
            "EXTRA_PYTHONPATH": ":".join(sys.path),
            "IGNORE_ENVIRONMENTS": "1",
        }
        exec_env.pop("PYTHONPATH", None)

        # noinspection PyTypeChecker
        os.execle(*exec_args, exec_env)


@dataclasses.dataclass
class DockerContainerEnvironment(ShellCommandEnvironmentAbs):
    """
    Each command is run inside a separate container.
    Supported since the 'ente' distribution.
    """
    image: str
    configuration: dict = dataclasses.field(default_factory=dict)

    def execute(self, shell, args: List[str]):
        from .shell import DTShell
        shell: DTShell
        # ---
        # TODO: implement this
        raise NotImplementedError("TODO")


DEFAULT_COMMAND_ENVIRONMENT: ShellCommandEnvironmentAbs = Python3Environment()



import os
import time
import copy
import glob
import traceback
import argparse
import dataclasses
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from types import SimpleNamespace
from typing import Dict, Type, Union, Optional, Mapping, Tuple, List, Any

from .repository import CommandsRepository
from .autocomplete import ArgumentParserCompleter
from .. import __version__, logger
from ..constants import CHECK_CMDS_UPDATE_MINS, DB_COMMAND_SET_UPDATES_CHECK, DTShellConstants, \
    EMBEDDED_COMMAND_SET_NAME
from ..environments import ShellCommandEnvironmentAbs, Python3Environment
from ..exceptions import UserError, InvalidRemote, CommandsLoadingException, CommandNotFound
from ..utils import run_cmd, undo_replace_spaces
from ..typing import DTShell

CommandName = str
CommandsTree = Dict[CommandName, Union[Mapping[CommandName, dict], Type['DTCommandAbs']]]


class DTCommandAbs(metaclass=ABCMeta):
    name: str = None
    level: int = None
    help: str = None
    parser: Optional[argparse.ArgumentParser] = None
    commands: CommandsTree = None
    descriptor: 'CommandDescriptor' = None
    fake: bool = False

    @staticmethod
    @abstractmethod
    def command(shell: DTShell, args: List[str]):
        """
        This function will be invoked when the user presses [Return] and runs the command.

        Parameters
        ----------
        shell : DTShell
            The instance of DTShell hosting this command.
        args : list
            A list of arguments passed to the command.
        """
        pass

    @staticmethod
    def complete(shell: DTShell, word: str, line: str):
        """
        This function will be invoked when the user presses the [Tab] key for auto completion.

        Parameters
        ----------
        shell : DTShell
            The instance of DTShell hosting this command.
        word : str
            The right-most word typed in the terminal
            (usually the string the user is trying to auto-complete).

        Returns
        -------
        list
            A list of strings. Each string is a suggestion for the user
            to auto-complete the command.
        """
        return []

    @staticmethod
    def fail(msg: str):
        raise Exception(msg)

    @classmethod
    def aliases(cls) -> List[str]:
        if not cls.descriptor:
            return []
        if not cls.descriptor.configuration:
            return []
        return cls.descriptor.configuration.aliases()

    @classmethod
    def get_command(cls, shell: DTShell, line: str) -> Tuple['CommandDescriptor', List[str]]:
        # print('>[%s]@(%s, %s)' % (line, cls.name, cls.__class__))
        line = line.strip()
        parts = [p.strip() for p in line.split(" ")]
        args = [p for p in parts if len(p) > 0]
        args = list(map(undo_replace_spaces, args))
        word = parts[0]
        # print('[%s, %r]@(%s, %s)' % (word, parts, cls.name, cls.__class__))
        if len(word) > 0:
            if len(cls.commands) > 0:
                # search every command and their aliases
                subcmds: Dict[str, Type[DTCommandAbs]] = copy.copy(cls.commands)
                for subcmd_name, subcmd in cls.commands.items():
                    subcmds.update({k: subcmd for k in subcmd.aliases()})
                # if we have a match we keep looking down recursively
                if word in subcmds:
                    return subcmds[word].get_command(shell, " ".join(parts[1:]))
                else:
                    raise CommandNotFound(last_matched=cls, remaining=parts)
            else:
                return cls.descriptor, args
        else:
            if len(cls.commands) > 0:
                raise CommandNotFound(last_matched=cls, remaining=parts)
            else:
                return cls.descriptor, args

    @classmethod
    def _complete(cls, shell: DTShell, word: str, line: str) -> List[str]:
        # start with user suggestions (i.e., implemented via DTCommand.complete())
        suggestions: List[str] = cls.complete(shell, word, line)
        if cls.parser is not None:
            completer: ArgumentParserCompleter = ArgumentParserCompleter(cls.parser)
            # add parser suggestions
            suggestions.extend(completer.get_completions(line))
        # ---
        return suggestions

    @classmethod
    def do_command(cls, shell: DTShell, line: str):
        descriptor: Optional[CommandDescriptor]
        args: List[str]
        # find the subcommand to execute
        descriptor, args = cls.get_command(shell, line)
        if descriptor is not None and not descriptor.command.fake:
            # annotate event
            shell.profile.events.new(
                "shell/command/execute",
                {"command_set": descriptor.command_set.as_dict(), "command": descriptor.selector}
            )
            # run command implementation
            return descriptor.command.command(shell, args)

    @classmethod
    def complete_command(cls, shell: DTShell, word: str, line: str, start_index: int, end_index: int) \
            -> List[str]:
        # TODO: add aliases
        subcmds = cls.commands.keys()
        parts = [p.strip() for p in line.split(" ")]
        partial_word: bool = len(word) != 0
        # NOTE: DEBUG only
        # logger.info(
        #     f"""
        #     line: |{line}|
        #     word: |{word}|
        #     start_index: {start_index}
        #     end_index: {end_index}
        #     partial_word: |{partial_word}|
        #     command: |{cls.name}|
        #     subcmds: |{subcmds}|
        #     parts: {parts}
        #     """
        # )
        # first word must match this command name
        if parts[0] == cls.name:
            # either there is only one word to complete or a full word and a partial word
            if len(parts) in [1, 2]:
                # strip this command name from the line
                nline: str = " ".join(parts[1:]) if len(parts) > 1 else line
                # collect all matching suggestions returned by the method DTCommand.complete()
                static_comp = [
                    k for k in cls._complete(shell, word, nline) if (not partial_word or k.startswith(word))
                ]
                # add all subcommands whose name match the word
                comp_subcmds = static_comp + [k for k in subcmds if (not partial_word or k.startswith(word))]
                return comp_subcmds
            # if we have that the first word matches the name of a subcommand, we pass the ball downstream
            if len(parts) > 1 and parts[1] in subcmds:
                child = parts[1]
                nline: str = " ".join(parts[1:])
                # let the child command autocomplete
                return cls.commands[child].complete_command(shell, word, nline, start_index, end_index)
            # we have a more complex partial line
            if len(parts) >= 2:
                # strip this command name from the line
                nline: str = " ".join(parts[1:])
                # collect all matching suggestions returned by the method DTCommand.complete()
                static_comp = [
                    k for k in cls._complete(shell, word, nline)
                ]
                return static_comp
        # ---
        return []

    @classmethod
    def help_command(cls, shell: DTShell):
        msg = cls.help if (cls.level == 0 and cls.help is not None) else str(shell.nohelp % cls.name)
        print(msg)


class DTCommandPlaceholder(DTCommandAbs):
    fake = True

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        return


class NoOpCommand(DTCommandAbs):
    @staticmethod
    def command(shell: DTShell, args: List[str], **kwargs):
        pass


class FailedToLoadCommand(NoOpCommand):
    @staticmethod
    def command(shell: DTShell, args: List[str], **kwargs):
        logger.warning("This command was not loaded")


class DTCommandConfigurationAbs(metaclass=ABCMeta):

    @classmethod
    def environment(cls, *args, **kwargs) -> Optional[ShellCommandEnvironmentAbs]:
        """
        The environment in which this command will run.
        """
        return None

    @classmethod
    def parser(cls, *args, **kwargs) -> Optional[argparse.ArgumentParser]:
        """
        The parser this command will use.
        """
        return None

    @classmethod
    def aliases(cls) -> List[str]:
        """
        Alternative names for this command.
        """
        return []


class DTCommandConfigurationDefault(DTCommandConfigurationAbs):
    pass


class DTCommandSetConfigurationAbs(metaclass=ABCMeta):
    path: str = None

    @classmethod
    def default_environment(cls, *args, **kwargs) -> Optional[ShellCommandEnvironmentAbs]:
        """
        The environment in which the commands in this set will run.
        """
        return None

    # noinspection PyUnusedLocal
    @classmethod
    def requirements(cls, *args, **kwargs) -> Optional[str]:
        """
        File containing the list of dependency python projects needed by the commands in this command set.
        """
        # no path => no requirements file
        if cls.path is None:
            return None
        # return the path to the requirements file if it exists
        command_set_metadir: str = os.path.join(os.path.abspath(cls.path), "__command_set__")
        requirements_fpath: str = os.path.join(command_set_metadir, "requirements.txt")
        return requirements_fpath if os.path.exists(requirements_fpath) else None

    @classmethod
    @abstractmethod
    def version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        """
        Version of this command set in the format (major, minor, patch).

        :return:    A tuple of three integer numbers representing (major, minor, path)
        """
        raise NotImplementedError("Subclasses of 'DTCommandSetConfigurationAbs' must implement the function "
                                  "version().")

    @classmethod
    @abstractmethod
    def minimum_shell_version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        """
        The minimum version of the shell neeeded for this command set to work properly.

        :return:    A tuple of three integer numbers representing (major, minor, path)
        """
        raise NotImplementedError("Subclasses of 'DTCommandSetConfigurationAbs' must implement the function "
                                  "minimum_shell_version().")

    @classmethod
    @abstractmethod
    def maximum_shell_version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        """
        The maximum version of the shell neeeded for this command set to work properly.

        :return:    A tuple of three integer numbers representing (major, minor, path)
        """
        raise NotImplementedError("Subclasses of 'DTCommandSetConfigurationAbs' must implement the function "
                                  "maximum_shell_version().")


class DTCommandSetConfigurationDefault(DTCommandSetConfigurationAbs):

    @classmethod
    def default_environment(cls, *args, **kwargs) -> Optional[ShellCommandEnvironmentAbs]:
        """
        The environment in which commands from this command set will run.
        """
        return Python3Environment()

    @classmethod
    def version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        return 0, 0, 0

    @classmethod
    def minimum_shell_version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        return 0, 0, 0

    @classmethod
    def maximum_shell_version(cls, *args, **kwargs) -> Tuple[int, int, int]:
        return 99, 99, 99


@dataclass
class CommandDescriptor:
    name: str
    path: str
    selector: str
    command_set: 'CommandSet'
    configuration: Type[DTCommandConfigurationAbs]
    environment: Optional[ShellCommandEnvironmentAbs] = None
    command: Type[DTCommandAbs] = None

    @property
    def aliases(self) -> List[str]:
        return self.configuration.aliases()


noop_command = SimpleNamespace(DTCommand=NoOpCommand)
failed_to_load_command = SimpleNamespace(DTCommand=FailedToLoadCommand)
default_command_configuration = SimpleNamespace(DTCommandConfiguration=DTCommandConfigurationDefault)
default_commandset_configuration = SimpleNamespace(DTCommandSetConfiguration=DTCommandSetConfigurationDefault)


@dataclass
class CommandSet:
    name: str
    path: str
    profile: Any
    repository: Optional[CommandsRepository] = None
    leave_alone: bool = False
    configuration: Type[DTCommandSetConfigurationAbs] = None
    commands: CommandsTree = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        from .importer import import_commandset_configuration
        # load command set configuration
        self.configuration: Type[DTCommandSetConfigurationAbs] = import_commandset_configuration(self)
        # load commands
        self.commands = self._find_commands()

    def init(self):
        from .importer import import_commandset_init
        # load command set init script
        import_commandset_init(self)

    @property
    def version(self) -> Optional[str]:
        # embedded command set
        if self.name == EMBEDDED_COMMAND_SET_NAME:
            return __version__
        # repository-based command sets
        if self.repository:
            return CommandsRepository.head_tag(self.path)
        # no repository
        return None

    @property
    def closest_version(self) -> Optional[str]:
        # embedded command set
        if self.name == EMBEDDED_COMMAND_SET_NAME:
            return __version__
        # repository-based command sets
        if self.repository:
            return CommandsRepository.closest_tag(self.path)
        # no repository
        return None

    @property
    def local_sha(self) -> Optional[str]:
        if self.repository is not None:
            stdout: str = run_cmd(["git", "-C", self.path, "rev-parse", "HEAD"])
            # noinspection PyTypeChecker
            return next(filter(len, stdout.split("\n")))

    def as_dict(self) -> dict:
        return {
            "name": self.name,
            "version": self.version,
            "closest_version": self.closest_version,
            "local_sha": self.local_sha,
            "repository": self.repository.as_dict() if self.repository is not None else None
        }

    def refresh(self):
        # reload commands
        self.commands = self._find_commands()

    def command_path(self, selector: str) -> str:
        return os.path.join(self.path, selector.strip(".").replace(".", os.path.sep))

    def download(self) -> bool:
        """Raises InvalidRemote if it cannot find it"""
        if self.repository is None:
            raise RuntimeError("You cannot 'download' a command set without a repository defined.")
        # ---
        remote_url = self.repository.remoteurl
        try:
            logger.info(f"Downloading Duckietown shell commands in {self.path} ...")
            # clone the repo
            branch: List[str] = ["--branch", self.repository.branch] if self.repository.branch else []
            run_cmd(["git", "clone"] + branch + ["--recurse-submodules", remote_url, self.path])
            logger.info(f"Commands downloaded successfully!")
            # refresh commands
            self.refresh()
        except Exception as e:
            # Excepts as InvalidRemote
            logger.error(f"Unable to clone the repo at '{remote_url}':\n{str(e)}.")
            return False

    def update(self) -> bool:
        # check that the repo is initialized in the commands path
        self.ensure_commands_exist()
        # update the commands if they are outdated
        return self.ensure_commands_updated()

    def ensure_commands_exist(self):
        # clone the commands if necessary
        if not os.path.exists(self.path):
            msg = f"I cannot find the command path {self.path}"
            if self.leave_alone:
                raise Exception(msg)
            logger.debug(msg)
            # we can download
            try:
                self.download()
            except InvalidRemote as e:
                msg = "I could not initialize the commands."
                raise CommandsLoadingException(msg) from e
        # make sure the commands exist
        if not os.path.exists(self.path):
            raise UserError(f"Commands not found at '{self.path}'.")

    def commands_need_update(self) -> bool:
        # command sets without repository cannot be updated
        if self.repository is None:
            return False
        # ---
        need_update: bool = False
        # get the current repo info
        db = self.profile.database(DB_COMMAND_SET_UPDATES_CHECK)
        # check if it's time to check for an update
        record: dict
        if db.contains(self.name):
            record = db.get(self.name)
            now = time.time()
            last_time_checked = record["time"]
            use_cached_commands = now - last_time_checked < CHECK_CMDS_UPDATE_MINS * 60
        else:
            # save the initial update record
            self.mark_as_just_updated()
            return False

        # check for an updated remote
        if not use_cached_commands:
            logger.info(f"Checking for updates for the command set '{self.name}'...")
            # get the local sha from file
            local_sha: Optional[str] = record["sha"]
            if local_sha is None:
                logger.error(f"Command set '{self.name}' has a repository but no local sha. "
                             f"This should not have happened. Contact technical support.")
                # TODO: maybe corrupted repository? suggest removing and reinstalling the command set?
                return False

            # get the remote sha from GitHub
            remote_sha: Optional[str] = self.repository.remote_sha()
            if remote_sha is None:
                return False

            # check if we need to update
            need_update = local_sha != remote_sha
            # touch flag to reset update check time
            self.mark_as_just_updated()
        # ---
        return need_update

    def mark_as_just_updated(self):
        db = self.profile.database(DB_COMMAND_SET_UPDATES_CHECK)
        db.set(self.name, {"sha": self.local_sha, "time": time.time()})

    def ensure_commands_updated(self) -> bool:
        # make sure the commands directory exists
        if not os.path.exists(self.path) and os.path.isdir(self.path):
            raise RuntimeError(f"There is no existing commands directory in '{self.path}'.")

        # command sets without repository cannot be updated
        if self.repository is None:
            raise RuntimeError("Command sets without a repository defined cannot be updated.")

        # Check for shell commands repo updates
        logger.debug(f"Checking for updates for the command set '{self.name}'...")
        if self.commands_need_update():
            logger.info(f"The command set '{self.name}' has available updates. Attempting to pull them.")
            wait_on_retry_secs = 4
            th = {2: "nd", 3: "rd", 4: "th"}
            for trial in range(3):
                try:
                    run_cmd(["git", "-C", self.path, "fetch", "origin", self.repository.branch])
                    run_cmd(["git", "-C", self.path, "reset", "--hard", f"origin/{self.repository.branch}"])
                    run_cmd(["git", "-C", self.path, "pull", "--recurse-submodules", "origin",
                             self.repository.branch])
                    logger.info(f"Command set '{self.name}' successfully updated!")
                except RuntimeError:
                    if DTShellConstants.VERBOSE:
                        traceback.print_exc()
                    logger.warning(
                        f"An error occurred while pulling the updated commands. In {wait_on_retry_secs} "
                        f"seconds we will retry for the {trial + 2}-{th[trial + 2]} time"
                    )
                    time.sleep(wait_on_retry_secs)
                else:
                    break
            run_cmd(["git", "-C", self.path, "submodule", "update"])
            # mark as updated
            self.mark_as_just_updated()
            # refresh commands
            self.refresh()
            # ---
            return True
        else:
            logger.debug(f"Command set '{self.name}' is up-to-date.")
            # ---
            return False

    def _find_commands(self, lvl: int = 0, all_commands: bool = False, selector: str = "",
                       path: Optional[str] = None) \
            -> Union[None, Dict[str, Union[dict, CommandDescriptor]], CommandDescriptor]:
        path: str = path or self.path
        entries = glob.glob(os.path.join(path, "*"))
        files = [os.path.basename(e) for e in entries if os.path.isfile(e)]
        dirs = [e for e in entries if os.path.isdir(e) and (lvl > 0 or os.path.basename(e) != "lib")]
        # base case: empty dir -> not a command
        if "command.py" not in files and not dirs:
            return None
        # commands that are not installed
        name: str = os.path.basename(path)
        # load subcommands
        subcmds = {}
        for d in dirs:
            cmd_name: str = os.path.basename(d)
            cmd_package = f"{selector}.{cmd_name}".lstrip(".")
            f = self._find_commands(lvl + 1, all_commands, cmd_package, d)
            if f is not None:
                subcmds[cmd_name] = f
        # not an empty directory, but not a command and not a container of subcommands either
        if "command.py" not in files and not subcmds:
            return None
        # leaf command
        if "command.py" in files and not subcmds:
            return CommandDescriptor(
                name=name,
                path=path,
                selector=selector,
                command_set=self,
                configuration=DTCommandConfigurationDefault,
                environment=None
            )
        # ---
        return subcmds



import grp
import os
import subprocess
import sys
from typing import List, Optional

from shutil import which

from ..exceptions import InvalidEnvironment, UserError
from .. import logger


def running_with_sudo() -> bool:
    return os.geteuid() == 0


def abort_if_running_with_sudo() -> None:
    if running_with_sudo():
        if "CIRCLECI" in os.environ:
            return
        msg = """\
Do not run dts using "sudo".'

As a matter of fact, do not run anything with "sudo" unless instructed to do so.\
"""
        raise UserError(msg)


def check_docker_environment():
    """Returns docker client"""
    check_executable_exists("docker")

    check_user_in_docker_group()
    #
    # if on_linux():
    #
    #     if username != 'root':
    #         check_user_in_docker_group()
    #     # print('checked groups')
    # else:
    #     logger.debug('skipping env check because not on Linux')

    try:
        import docker
    except Exception as e:
        msg = "Could not import package docker:\n%s" % e
        msg += "\n\nYou need to install the package"
        raise InvalidEnvironment(msg)

    if "DOCKER_HOST" in os.environ:
        msg = 'Note that the variable DOCKER_HOST is set to "%s"' % os.environ["DOCKER_HOST"]
        logger.warning(msg)

    try:
        # noinspection PyUnresolvedReferences
        client = docker.from_env()

    except Exception as e:
        msg = "I cannot communicate with Docker:\n%s" % e
        msg += "\n\nMake sure the docker service is running."
        raise InvalidEnvironment(msg)

    return client


def on_linux() -> bool:
    return sys.platform.startswith("linux")


def check_executable_exists(cmdname: str) -> None:
    p = which(cmdname)
    if p is None:
        msg = 'Could not find executable "%s".' % cmdname
        raise InvalidEnvironment(msg)


def check_user_in_docker_group() -> None:
    # first, let's see if there exists a group "docker"
    group_names = [g.gr_name for g in grp.getgrall()]
    G = "docker"
    if G not in group_names:
        pass
    else:
        group_id = grp.getgrnam(G).gr_gid
        my_groups = os.getgroups()
        if group_id not in my_groups:
            msg = 'My groups are %s and "%s" group is %s ' % (my_groups, G, group_id)
            msg += "\n\nNote that when you add a user to a group, you need to login in and out."
            # logger.debug(msg)


def get_active_groups(username: Optional[str] = None) -> List[str]:
    cmd = ["groups"]

    if username:
        cmd.append(username)

    try:
        stdout = subprocess.check_output(["groups"])
    except subprocess.CalledProcessError as e:
        raise Exception(str(e))
    active_groups = stdout.decode().split()

    return active_groups



import os
from typing import Optional

from dt_shell_cli import logger
from .database import DTShellDatabase
from ..constants import DTShellConstants
from ..utils import indent_block


class InstalledDependenciesDatabase(DTShellDatabase):

    @classmethod
    def load(cls, profile):
        from ..profile import ShellProfile
        profile: ShellProfile
        # ---
        name: str = "installed_dependencies"
        return profile.database(name, cls=InstalledDependenciesDatabase)

    def contains(self, dependencies_fpath: Optional[str]) -> bool:
        # no dependencies filepath? => do no have it
        if dependencies_fpath is None:
            return False
        dependencies_fpath: str = os.path.abspath(dependencies_fpath)
        return super(InstalledDependenciesDatabase, self).contains(dependencies_fpath)

    def needs_install_step(self, dependencies_fpath: Optional[str]) -> bool:
        # no dependencies files => no need to install
        if dependencies_fpath is None:
            return False
        dependencies_fpath: str = os.path.abspath(dependencies_fpath)
        # get current list
        with open(dependencies_fpath, "rt") as fin:
            current: str = fin.read()
        # get list from last installation
        last: str = self.get(dependencies_fpath, "")
        # compare the two
        if DTShellConstants.VERBOSE:
            logger.debug(f"Comparing OLD<>NEW dependencies lists for '{dependencies_fpath}':\n\n"
                         f"OLD:\n|{indent_block(last)}|\n\nNEW:\n|{indent_block(current)}|")
        return current != last

    def mark_as_installed(self, dependencies_fpath: Optional[str]):
        # no dependencies files no need to install
        if dependencies_fpath is None:
            return
        dependencies_fpath: str = os.path.abspath(dependencies_fpath)
        # get current list
        with open(dependencies_fpath, "rt") as fin:
            current: str = fin.read()
        # set list as last installation
        self.set(dependencies_fpath, current)



import sys
from typing import Optional, List

from dt_shell import DTCommandAbs, DTShell, dtslogger
from dt_shell.constants import SHELL_REQUIREMENTS_LIST
from dt_shell.database.utils import InstalledDependenciesDatabase
from dt_shell.utils import pip_install


class DTCommand(DTCommandAbs):

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # install dependencies
        cache: InstalledDependenciesDatabase = InstalledDependenciesDatabase.load(shell.profile)
        # - shell
        dtslogger.info("Installing shell dependencies...")
        pip_install(sys.executable, SHELL_REQUIREMENTS_LIST)
        cache.mark_as_installed(SHELL_REQUIREMENTS_LIST)
        # - command sets
        for cs in shell.command_sets:
            requirements_list: Optional[str] = cs.configuration.requirements()
            dtslogger.info(f"Installing dependencies for command set '{cs.name}'...")
            pip_install(sys.executable, requirements_list)
            cache.mark_as_installed(requirements_list)



import argparse
from typing import Optional, List

from dt_shell.commands import DTCommandConfigurationAbs
from dt_shell.environments import Python3Environment, ShellCommandEnvironmentAbs


class DTCommandConfiguration(DTCommandConfigurationAbs):

    @classmethod
    def environment(cls, **kwargs) -> Optional[ShellCommandEnvironmentAbs]:
        return Python3Environment()

    @classmethod
    def parser(cls, **kwargs) -> Optional[argparse.ArgumentParser]:
        pass

    @classmethod
    def aliases(cls) -> List[str]:
        return ["quit"]



import json
import traceback
from abc import abstractmethod
from threading import Thread
from typing import Optional

import requests
from dt_shell.utils import DebugInfo

import dockertown
from dt_shell.database import DTShellDatabase

from dt_shell_cli import logger
from .constants import DTShellConstants, DTHUB_URL, DB_BILLBOARDS
from .hub import HUBApiError, hub_api_post
from .shell import Event, DTShell
from .statistics import ShellProfileEventsDatabase


class Task(Thread):

    def __init__(self, shell: DTShell, name: str, killable: bool = False, **kwargs):
        super(Task, self).__init__(daemon=True, **kwargs)
        # arguments
        self._shell: DTShell = shell
        self._name: str = name
        self._killable: bool = killable
        # internal state
        self._has_started: bool = False
        self._has_finished: bool = False
        # register shutdown handlers
        self._requested_shutdown: bool = False
        self._shell.on_keyboard_interrupt(self._shutdown)
        self._shell.on_shutdown(self._shutdown)

    @property
    def name(self) -> str:
        return self._name

    @property
    def killable(self) -> bool:
        return self._killable

    @property
    def started(self) -> bool:
        return self._has_started

    @property
    def finished(self) -> bool:
        return self._has_finished

    def run(self) -> None:
        # mark as started
        if DTShellConstants.VERBOSE:
            logger.debug(f"Task '{self._name}' started!")
        self._has_started = True
        # execute task job
        try:
            self.execute()
        except KeyboardInterrupt:
            logger.debug(f"Task '{self._name}' interrupted by SIGINT!")
        finally:
            # mark as finished
            self._has_finished = True
        if DTShellConstants.VERBOSE:
            logger.debug(f"Task '{self._name}' finished!")

    def _shutdown(self, event: Event) -> None:
        # prevent multiple shutdown requests
        if self._requested_shutdown:
            return
        self._requested_shutdown = True
        # forward shutdown request to job implementation
        if DTShellConstants.VERBOSE:
            logger.debug(f"Stopping task '{self._name}'...")
        self.shutdown(event)
        # wait for the task to finish if we marked it as not killable
        if not self._killable:
            if DTShellConstants.VERBOSE:
                logger.debug(f"Waiting for task '{self._name}' to finish...")
            self.join()

    @abstractmethod
    def execute(self):
        pass

    @abstractmethod
    def shutdown(self, event: Event):
        pass


class UpdateBillboardsTask(Task):

    def __init__(self, shell, **kwargs):
        super(UpdateBillboardsTask, self).__init__(shell, name="billboards-updater", **kwargs)
        self._db: DTShellDatabase = DTShellDatabase.open(DB_BILLBOARDS)

    def execute(self):
        url: str = f"{DTHUB_URL}/api/v1/shell/billboards"
        # reach out to the HUB and grub the new billboards
        try:
            response: dict = requests.get(url).json()
            self._shell.profile.events.new("shell/billboards/update")
        except:
            logger.warning("An error occurred while updating the billboards")
            logger.debug(traceback.format_exc())
            # mark as updated so we don't retry right away
            self._shell.mark_updated("billboards")
            return
        # check response
        if response.get("success", False) is not True:
            logger.warning("An error occurred while updating the billboards")
            logger.debug("HUB response:\n" + json.dumps(response, indent=4, sort_keys=True))
            # mark as updated so we don't retry right away
            self._shell.mark_updated("billboards")
            return
        # update local database
        self._db.clear()
        self._db.update(response.get("result", {}))
        self._shell.mark_updated("billboards")
        logger.debug("Billboards updated!")

    def shutdown(self, event: Event):
        pass


class CollectDockerVersionTask(Task):

    def __init__(self, shell, **kwargs):
        super(CollectDockerVersionTask, self).__init__(shell, name="docker-version-probe", killable=True,
                                                       **kwargs)
        self._db: DTShellDatabase = DTShellDatabase.open(DB_BILLBOARDS)

    def execute(self):
        try:
            # create docker client
            docker = dockertown.DockerClient()
            # get versions
            versions: dict = docker.version()
            DebugInfo.name2versions["docker/client"] = versions["Client"]["Version"]
            DebugInfo.name2versions["docker/server"] = versions["Server"]["Version"]
            DebugInfo.name2versions["docker/client/api"] = versions["Client"]["ApiVersion"]
            DebugInfo.name2versions["docker/server/api"] = versions["Server"]["ApiVersion"]
        except:
            DebugInfo.name2versions["docker/client"] = "(error)"
            DebugInfo.name2versions["docker/server"] = "(error)"
            DebugInfo.name2versions["docker/client/api"] = "(error)"
            DebugInfo.name2versions["docker/server/api"] = "(error)"

    def shutdown(self, event: Event):
        pass


class UploadStatisticsTask(Task):

    def __init__(self, shell, **kwargs):
        super(UploadStatisticsTask, self).__init__(shell, name="stats-uploader", killable=True, **kwargs)
        self._db: ShellProfileEventsDatabase = shell.profile.events
        self._token: Optional[str] = shell.profile.secrets.dt_token

    def execute(self):
        if self._token is None:
            return
        # push events
        for evt in self._db.events():
            try:
                data: dict = {
                    "key": evt.name,
                    "format": evt.format,
                    "payload": evt.payload,
                    "stamp": evt.time_millis
                }
                if evt.labels:
                    data["labels"] = evt.labels
                hub_api_post("statistics/user/event", data)
                evt.delete()
                if DTShellConstants.VERBOSE:
                    logger.debug(f"Event '{evt.__key__}' pushed to the HUB")
            except HUBApiError as e:
                if e.code == 409:
                    # duplicated stats points
                    evt.delete()
                    continue
                if DTShellConstants.VERBOSE:
                    logger.debug(e.human)
            except:
                if DTShellConstants.VERBOSE:
                    logger.debug(traceback.format_exc())
        # mark as done
        self._shell.mark_done("upload_events")

    def shutdown(self, event: Event):
        pass




# -*- coding: utf-8 -*-
import logging
import pathlib
import warnings

warnings.filterwarnings(action='ignore', module='.*paramiko.*')

# ===> COMPATIBILITY
# these are only for older versions of the commands, we will drop these with the shell v7
COMPATIBILITY_DIR = str(pathlib.Path(__file__).parent.joinpath("compatibility").resolve())
__path__ += [COMPATIBILITY_DIR]

from .compatibility import duckietown_tokens
from .compatibility import OtherVersions
from .compatibility.env_checks import check_package_version
# <=== COMPATIBILITY


FORMAT = "%(levelname)s:%(name)s : %(message)s"
logging.basicConfig(format=FORMAT)

from dt_shell_cli import logger

# logger dedicated to the commands
dtslogger = logging.getLogger("dts")
dtslogger.setLevel(logging.INFO)

__version__ = "6.0.3"

import sys

if sys.version_info < (3, 6):
    msg = f"! duckietown-shell works with Python 3.6 and later !.\nDetected {sys.version}."
    logger.error(msg)
    sys.exit(2)

from .shell import DTShell

from .commands import DTCommandAbs, DTCommandPlaceholder
from .exceptions import *
from typing import Optional

# singleton
shell: Optional[DTShell] = None



# used by the following commands files:
#
#   - commands/command.py
#   - install/command.py
#
from typing import Optional, Dict


def _get_commands(*args, **kwargs) -> Optional[Dict[str, object]]:
    import dt_shell
    return dt_shell.shell.commands
